# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("AggregatorConfig",)


import pydantic


class AggregatorConfig(pydantic.BaseModel):
    """Configuration for the provenance aggregator."""

    output_path: str | None = None
    """Path for the output provenance quantum graph file.

    At present this option is intended only for debugging.
    """

    worker_log_dir: str | None = None
    """Path to a directory (POSIX only) for parallel worker logs."""

    worker_log_level: str = "VERBOSE"
    """Log level for worker processes/threads.

    Per-quantum messages only appear at ``DEBUG`` level.
    """

    worker_profile_dir: str | None = None
    """Path to a directory (POSIX only) for parallel worker profiling dumps.

    This option is ignored when `n_processes` is `1`.
    """

    n_processes: int = 1
    """Number of processes the scanner should use."""

    assume_complete: bool = True
    """If `True`, the aggregator can assume all quanta have run to completion
    (including any automatic retries).  If `False`, only successes can be
    considered final, and quanta that appear to have failed or to have not been
    executed are ignored.
    """

    defensive_ingest: bool = False
    """If `True`, guard against datasets having already been ingested into the
    central butler repository.

    Defensive ingest mode is automatically turned on (with a warning emitted)
    if an ingest attempt fails due to a database constraint violation. Enabling
    defensive mode up-front avoids this warning and is slightly more efficient
    when it is already known that some datasets have already been ingested.

    Defensive mode does not guard against race conditions from multiple ingest
    processes running simultaneously, as it relies on a one-time query to
    determine what is already present in the central repository.
    """

    ingest_batch_size: int = 10000
    """Number of butler datasets that must accumulate to trigger an ingest."""

    register_dataset_types: bool = True
    """Whether to register output dataset types in the central butler
    repository before starting ingest.
    """

    update_output_chain: bool = True
    """Whether to prepend the output `~lsst.daf.butler.CollectionType.RUN` to
    the output `~lsst.daf.butler.CollectionType.CHAINED` collection.
    """

    dry_run: bool = False
    """If `True`, do not actually perform any deletions or central butler
    ingests.

    Most log messages concerning deletions and ingests will still be emitted in
    order to provide a better emulation of a real run.
    """

    interactive_status: bool = False
    """Whether to use an interactive status display with progress bars.

    If this is `True`, the `tqdm` module must be available.  If this is
    `False`, a periodic logger will be used to display status at a fixed
    interval instead (see `log_status_interval`).
    """

    log_status_interval: float | None = None
    """Interval (in seconds) between periodic logger status updates."""

    worker_sleep: float = 0.01
    """Time (in seconds) a worker should wait when there are no requests from
    the main aggregator process.
    """

    zstd_level: int = 10
    """ZStandard compression level to use for all compressed-JSON blocks."""

    zstd_dict_size: int = 32768
    """Size (in bytes) of the ZStandard compression dictionary."""

    zstd_dict_n_inputs: int = 512
    """Number of samples of each type (see below) to include in ZStandard
    compression dictionary training.

    Training is run on a random subset of the `PredictedQuantumDatasetsModel`
    objects in the predicted graph, as well as the first provenance quanta,
    logs, and metadata blocks encountered.
    """

    mock_storage_classes: bool = False
    """Enable support for storage classes by created by the
    lsst.pipe.base.tests.mocks package.
    """



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import dataclasses
import itertools
import os
import time
import unittest.mock
import uuid
from collections.abc import Iterator
from contextlib import contextmanager

import numpy as np
from click.testing import CliRunner, Result

import lsst.utils.tests
from lsst.daf.butler import Butler, ButlerLogRecords, QuantumBackedButler
from lsst.pipe.base import AlgorithmError, QuantumSuccessCaveats, TaskMetadata
from lsst.pipe.base import automatic_connection_constants as acc
from lsst.pipe.base.cli.cmd.commands import aggregate_graph as aggregate_graph_cli
from lsst.pipe.base.graph_walker import GraphWalker
from lsst.pipe.base.pipeline_graph import Edge
from lsst.pipe.base.quantum_graph import (
    PredictedDatasetInfo,
    PredictedQuantumGraph,
    PredictedQuantumInfo,
    ProvenanceDatasetInfo,
    ProvenanceQuantumGraph,
    ProvenanceQuantumGraphReader,
    ProvenanceQuantumInfo,
)
from lsst.pipe.base.quantum_graph.aggregator import AggregatorConfig, FatalWorkerError, aggregate_graph
from lsst.pipe.base.quantum_provenance_graph import QuantumRunStatus
from lsst.pipe.base.resource_usage import QuantumResourceUsage
from lsst.pipe.base.single_quantum_executor import SingleQuantumExecutor
from lsst.pipe.base.tests.mocks import (
    DirectButlerRepo,
    DynamicConnectionConfig,
    DynamicTestPipelineTaskConfig,
)
from lsst.pipe.base.tests.util import patch_deterministic_uuid4
from lsst.resources import ResourcePath
from lsst.utils.packages import Packages


@dataclasses.dataclass
class PrepInfo:
    """Struct of objects used in an aggregator test."""

    butler: Butler
    butler_path: str
    predicted: PredictedQuantumGraph
    predicted_path: str
    config: AggregatorConfig


class AggregatorTestCase(unittest.TestCase):
    """Unit tests for `lsst.pipe.base.quantum_graph.aggregator`."""

    @staticmethod
    @contextmanager
    def make_test_repo() -> Iterator[PrepInfo]:
        """Make a test data repository and predicted quantum graph.

        Returns
        -------
        prep_info : `PrepInfo`
            Objects used in aggregator tests.

        Notes
        -----
        The pipeline graph used by this task looks like this:

            ■  calibrate: {detector, visit}
          ╭─┤
          ■ │  consolidate: {visit}
            │
            ■  resample: {patch, visit}
            │
            ■  coadd: {band, patch}

        The data can be visualized via::

            python -m lsst.daf.butler.tests.registry_data.spatial

        One of the 'calibrate' quanta (visit=2, detector=2) is configured to
        fail with `lsst.pipe.base.AnnotatedPartialOutputsError`.  This lets us
        test both success-with-caveats and failures, depending on how we
        configure the executor.  This ``{visit: 2, detector: 2}`` data ID is
        the only one that overlaps ``{tract: 1, patch: 1}`` and
        ``{tract: 0, patch: 5}``, so it should chain to the 'resample' and
        'coadd' tasks, too.
        """
        with patch_deterministic_uuid4(100):
            with DirectButlerRepo.make_temporary("base.yaml", "spatial.yaml") as (helper, root):
                calibrate_config = DynamicTestPipelineTaskConfig()
                calibrate_config.fail_exception = "lsst.pipe.base.AnnotatedPartialOutputsError"
                calibrate_config.fail_condition = "visit=2 AND detector=2"
                helper.add_task(
                    "calibrate",
                    config=calibrate_config,
                    dimensions=["visit", "detector"],
                    inputs={
                        "input_image": DynamicConnectionConfig(
                            dataset_type_name="raw",
                            dimensions=["visit", "detector"],
                        )
                    },
                    prerequisite_inputs={
                        "refcat": DynamicConnectionConfig(
                            dataset_type_name="references",
                            dimensions=["htm7"],
                            multiple=True,
                        )
                    },
                    init_outputs={
                        "output_schema": DynamicConnectionConfig(
                            dataset_type_name="source_schema",
                        )
                    },
                    outputs={
                        "output_image": DynamicConnectionConfig(
                            dataset_type_name="image",
                            dimensions=["visit", "detector"],
                        ),
                        "output_table": DynamicConnectionConfig(
                            dataset_type_name="source_detector",
                            dimensions=["visit", "detector"],
                        ),
                    },
                )
                helper.add_task(
                    "consolidate",
                    dimensions=["visit"],
                    init_inputs={
                        "input_schema": DynamicConnectionConfig(
                            dataset_type_name="source_schema",
                        )
                    },
                    inputs={
                        "input_table": DynamicConnectionConfig(
                            dataset_type_name="source_detector",
                            dimensions=["visit", "detector"],
                            multiple=True,
                        )
                    },
                    outputs={
                        "output_table": DynamicConnectionConfig(
                            dataset_type_name="source",
                            dimensions=["visit"],
                        )
                    },
                )
                helper.add_task(
                    "resample",
                    dimensions=["patch", "visit"],
                    inputs={
                        "input_image": DynamicConnectionConfig(
                            dataset_type_name="image",
                            dimensions=["visit", "detector"],
                            multiple=True,
                        )
                    },
                    outputs={
                        "output_image": DynamicConnectionConfig(
                            dataset_type_name="warp",
                            dimensions=["patch", "visit"],
                        )
                    },
                )
                helper.add_task(
                    "coadd",
                    dimensions=["patch", "band"],
                    inputs={
                        "input_image": DynamicConnectionConfig(
                            dataset_type_name="warp",
                            dimensions=["patch", "visit"],
                            multiple=True,
                        )
                    },
                    outputs={
                        "output_image": DynamicConnectionConfig(
                            dataset_type_name="coadd",
                            dimensions=["patch", "band"],
                        ),
                    },
                )
                pqgc = helper.make_quantum_graph_builder().finish(output="out_chain")
                # We use the butler root for various QG files just because it's
                # a convenient temporary directory.
                predicted_path = os.path.join(root, "predicted.qg")
                pqgc.write(predicted_path)
                config = AggregatorConfig(
                    output_path=os.path.join(root, "provenance.qg"),
                    # Set these small to see logic paths that otherwise only
                    # affect large graphs.
                    ingest_batch_size=10,
                    zstd_dict_size=256,
                    zstd_dict_n_inputs=16,
                )
                yield PrepInfo(
                    butler=helper.butler,
                    butler_path=root,
                    predicted=pqgc.assemble(),
                    predicted_path=predicted_path,
                    config=config,
                )

    def iter_graph_execution(
        self, repo: ResourcePath, qg: PredictedQuantumGraph, raise_on_partial_outputs: bool
    ) -> Iterator[uuid.UUID]:
        """Return an iterator that executes and yields quanta one by one.

        Parameters
        ----------
        repo : `lsst.resources.ResourcePath`
            Butler repository path.
        qg : `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Predicted quantum graph.  Must have datastore records attached,
            since execution uses a quantum-backed butler.
        raise_on_partial_outputs : `bool`
            Whether to raise on `lsst.pipe.base.AnnotatedPartialOutputsError`
            or treat it as a success with caveats.

        Returns
        -------
        quanta : `~collections.abc.Iterator` [`uuid.UUID`]
            An iterator over successful quantum IDs.  Failed and blocked quanta
            are not included.
        """
        qg.init_output_run(qg.make_init_qbb(repo))
        sqe = SingleQuantumExecutor(
            limited_butler_factory=lambda quantum: QuantumBackedButler.initialize(
                repo,
                quantum,
                qg.pipeline_graph.universe,
            ),
            assume_no_existing_outputs=True,
            raise_on_partial_outputs=raise_on_partial_outputs,
        )
        qg.build_execution_quanta()
        xgraph = qg.quantum_only_xgraph
        walker = GraphWalker[uuid.UUID](xgraph.copy())
        for ready in walker:
            for quantum_id in ready:
                info = xgraph.nodes[quantum_id]
                try:
                    sqe.execute(info["pipeline_node"], info["quantum"], quantum_id)
                except AlgorithmError:
                    walker.fail(quantum_id)
                else:
                    walker.finish(quantum_id)
                    yield quantum_id

    def check_provenance_graph(
        self,
        pred: PredictedQuantumGraph,
        prov_reader: ProvenanceQuantumGraphReader,
        butler: Butler,
        expect_failure: bool,
        start_time: float,
    ) -> None:
        """Run a batter of tests on a provenance quantum graph produced by
        scanning the graph created by `make_test_repo`.

        Parameters
        ----------
        pred: `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Predicted quantum graph.
        prov_reader : \
                `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraphReader`
            Reader for the provenance quantum graph.
        butler : `lsst.daf.butler.Butler`
            Client for the data repository.
        expect_failure : `bool`
            Whether to expect one quantum of 'calibrate' to fail (`True`) or
            succeed without writing anything (`False`).
        start_time : `float`
            A POSIX timestamp that strictly precedes the start time of any
            quantum's execution.
        """
        prov_reader.read_full_graph()
        prov = prov_reader.graph
        checked_some_metadata = False
        checked_some_log = False
        self.maxDiff = None
        self.assertEqual(
            list(butler.collections.get_info(prov.header.output).children),
            [prov.header.output_run]
            + list(butler.collections.query(prov.header.inputs, flatten_chains=True)),
        )
        self.assertEqual(pred.quanta_by_task.keys(), prov.quanta_by_task.keys())
        for task_label in pred.quanta_by_task:
            self.assertEqual(pred.quanta_by_task[task_label], prov.quanta_by_task[task_label])
        self.assertEqual(pred.datasets_by_type.keys() - {"packages"}, prov.datasets_by_type.keys())
        for dataset_type_name in prov.datasets_by_type:
            self.assertEqual(
                pred.datasets_by_type[dataset_type_name], prov.datasets_by_type[dataset_type_name]
            )
        self.assertEqual(prov.init_quanta.keys(), pred.quanta_by_task.keys())
        for quantum_id in pred:
            # Check consistency between the predicted and provenance quantum
            # node attributes.
            pred_qinfo: PredictedQuantumInfo = pred.bipartite_xgraph.nodes[quantum_id]
            prov_qinfo: ProvenanceQuantumInfo = prov.bipartite_xgraph.nodes[quantum_id]
            self.assertEqual(pred_qinfo["task_label"], prov_qinfo["task_label"])
            self.assertEqual(pred_qinfo["data_id"], prov_qinfo["data_id"])
            msg = f"{pred_qinfo['task_label']}@{pred_qinfo['data_id']}"
            # Check consistency between the predicted and provenance dataset
            # node attributes and edges.  Also gather existence information for
            # use later.
            existence: dict[str, list[bool]] = {}
            pipeline_edges: list[Edge]
            for dataset_id, _, pipeline_edges in pred.bipartite_xgraph.in_edges(
                quantum_id, data="pipeline_edges"
            ):
                self.assertTrue(prov.bipartite_xgraph.has_predecessor(quantum_id, dataset_id))
                for edge in pipeline_edges:
                    existence.setdefault(edge.connection_name, []).append(
                        self.check_dataset(dataset_id, pred, prov, butler)
                    )
            for _, dataset_id, pipeline_edges in pred.bipartite_xgraph.out_edges(
                quantum_id, data="pipeline_edges"
            ):
                self.assertTrue(prov.bipartite_xgraph.has_successor(quantum_id, dataset_id))
                for edge in pipeline_edges:
                    existence.setdefault(edge.connection_name, []).append(
                        self.check_dataset(dataset_id, pred, prov, butler)
                    )
            # Check quantum status and dataset existence against the known
            # structure of the graph and where failures/caveats occur.
            match (pred_qinfo["task_label"], dict(pred_qinfo["data_id"].required)):
                case "calibrate", {"visit": 2, "detector": 2}:
                    # This is the quantum that can directly raise.
                    self._expect_all_exist(existence["input_image"], msg=msg)
                    self._expect_all_exist(existence["refcat"], msg=msg)
                    self._expect_none_exist(existence["output_image"], msg=msg)
                    self._expect_none_exist(existence["output_table"], msg=msg)
                    if expect_failure:
                        self._expect_failure(prov_qinfo, existence, msg=msg)
                    else:
                        self._expect_successful(
                            prov_qinfo,
                            existence,
                            caveats=(
                                QuantumSuccessCaveats.PARTIAL_OUTPUTS_ERROR
                                | QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
                                | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
                            ),
                            exception_type="lsst.pipe.base.tests.mocks.MockAlgorithmError",
                            msg=msg,
                        )
                case "consolidate", {"visit": 2}:
                    # This quantum will succeed (with one predicted input
                    # missing) or be blocked.
                    self._expect_one_missing(existence["input_table"], msg=msg)
                    if expect_failure:
                        self._expect_blocked(prov_qinfo, existence, msg=msg)
                    else:
                        self._expect_successful(prov_qinfo, existence, msg=msg)
                case (
                    "resample" | "coadd",
                    {"tract": 1, "patch": 1} | {"tract": 0, "patch": 5},
                ):
                    # These quanta will be blocked by an upstream failure or do
                    # chained caveats, since they won't have enough inputs to
                    # run.
                    if expect_failure:
                        self._expect_blocked(prov_qinfo, existence, msg=msg)
                    else:
                        self._expect_successful(
                            prov_qinfo,
                            existence,
                            caveats=(
                                QuantumSuccessCaveats.ADJUST_QUANTUM_RAISED
                                | QuantumSuccessCaveats.NO_WORK
                                | QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
                                | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
                            ),
                            msg=msg,
                        )
                case (
                    "resample",
                    {"tract": 0, "patch": 4, "visit": 2} | {"tract": 1, "patch": 0, "visit": 2},
                ):
                    # This will succeed or be blocked, with one input missing
                    # regardless.
                    self._expect_one_missing(existence["input_image"], msg=msg)
                    if expect_failure:
                        self._expect_blocked(prov_qinfo, existence, msg=msg)
                    else:
                        self._expect_successful(prov_qinfo, existence, msg=msg)
                case (
                    "coadd",
                    {"tract": 0, "patch": 4, "band": "r"} | {"tract": 1, "patch": 0, "band": "r"},
                ):
                    # This will succeed with no inputs missing or be blocked
                    # with one input missing.
                    if expect_failure:
                        self._expect_one_missing(existence["input_image"], msg=msg)
                        self._expect_blocked(prov_qinfo, existence, msg=msg)
                    else:
                        self._expect_all_exist(existence["input_image"], msg=msg)
                        self._expect_successful(prov_qinfo, existence, msg=msg)
                case _:
                    # All other quanta should succeed and have all inputs
                    # present.
                    for connection_name in prov_qinfo["pipeline_node"].inputs.keys():
                        self._expect_all_exist(existence[connection_name], msg=msg)
                    self._expect_successful(prov_qinfo, existence, msg=msg)
            if not checked_some_metadata and prov_qinfo["status"] is QuantumRunStatus.SUCCESSFUL:
                self.check_metadata(quantum_id, prov_reader, butler)
                checked_some_metadata = True
            if not checked_some_log and prov_qinfo["status"] in (
                QuantumRunStatus.SUCCESSFUL,
                QuantumRunStatus.FAILED,
            ):
                self.check_log(quantum_id, prov_reader, butler)
                checked_some_log = True
        self.assertTrue(checked_some_metadata)
        self.assertTrue(checked_some_log)
        self.check_resource_usage_table(
            prov_reader.graph, expect_failure=expect_failure, start_time=start_time
        )
        self.check_packages(prov_reader)
        self.check_quantum_table(prov_reader.graph, expect_failure=expect_failure)
        self.check_exception_table(prov_reader.graph, expect_failure=expect_failure)

    def _expect_all_exist(self, existence: list[bool], msg: str) -> None:
        self.assertTrue(all(existence), msg=msg)

    def _expect_none_exist(self, existence: list[bool], msg: str) -> None:
        self.assertFalse(any(existence), msg=msg)

    def _expect_one_missing(self, existence: list[bool], msg: str) -> None:
        self.assertEqual(existence.count(False), 1, msg=msg)

    def _expect_successful(
        self,
        info: ProvenanceQuantumInfo,
        existence: dict[str, list[bool]],
        caveats: QuantumSuccessCaveats = QuantumSuccessCaveats.NO_CAVEATS,
        exception_type: str | None = None,
        *,
        msg: str,
    ) -> None:
        self.assertEqual(info["status"], QuantumRunStatus.SUCCESSFUL, msg=msg)
        self.assertEqual(info["caveats"], caveats, msg=msg)
        if exception_type is None:
            self.assertIsNone(info["exception"], msg=msg)
        else:
            assert info["exception"] is not None
            self.assertEqual(info["exception"].type_name, exception_type, msg=msg)
        self._expect_all_exist(existence[acc.LOG_OUTPUT_CONNECTION_NAME], msg=msg)
        self._expect_all_exist(existence[acc.METADATA_OUTPUT_CONNECTION_NAME], msg=msg)
        if not (caveats & QuantumSuccessCaveats.ANY_OUTPUTS_MISSING):
            for connection_name in info["pipeline_node"].outputs.keys():
                self._expect_all_exist(existence[connection_name], msg=msg)
        if caveats & QuantumSuccessCaveats.ALL_OUTPUTS_MISSING:
            for connection_name in info["pipeline_node"].outputs.keys():
                self._expect_none_exist(existence[connection_name], msg=msg)
        self.assertIsNotNone(info["resource_usage"], msg=msg)
        self.assertGreater(info["resource_usage"].total_time, 0, msg=msg)
        self.assertGreater(info["resource_usage"].memory, 0, msg=msg)

    def _expect_failure(
        self, info: ProvenanceQuantumInfo, existence: dict[str, list[bool]], msg: str
    ) -> None:
        self.assertEqual(info["status"], QuantumRunStatus.FAILED, msg=msg)
        self._expect_all_exist(existence[acc.LOG_OUTPUT_CONNECTION_NAME], msg=msg)
        self._expect_none_exist(existence[acc.METADATA_OUTPUT_CONNECTION_NAME], msg=msg)
        for connection_name in info["pipeline_node"].outputs.keys():
            self._expect_none_exist(existence[connection_name], msg=msg)

    def _expect_blocked(
        self,
        info: ProvenanceQuantumInfo,
        existence: dict[str, list[bool]],
        msg: str,
    ) -> None:
        self.assertEqual(info["status"], QuantumRunStatus.BLOCKED, msg=msg)
        self._expect_none_exist(existence[acc.LOG_OUTPUT_CONNECTION_NAME], msg=msg)
        self._expect_none_exist(existence[acc.METADATA_OUTPUT_CONNECTION_NAME], msg=msg)
        for connection_name in info["pipeline_node"].outputs.keys():
            self._expect_none_exist(existence[connection_name], msg=msg)

    def check_dataset(
        self,
        dataset_id: uuid.UUID,
        pred: PredictedQuantumGraph,
        prov: ProvenanceQuantumGraph,
        butler: Butler,
    ) -> bool:
        """Check a provenance dataset for consistency with its predicted
        counterpart.

        Parameters
        ----------
        dataset_id : `uuid.UUID`
            Unique ID for the dataset.
        pred: `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Predicted quantum graph.
        prov : `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraph`
            Provenance quantum graph.
        butler : `lsst.daf.butler.Butler`
            Client for the data repository.

        Returns
        -------
        exists : `bool`
            Whether the dataset was marked as existing in the provenance
            quantum graph.
        """
        pred_info: PredictedDatasetInfo = pred.bipartite_xgraph.nodes[dataset_id]
        prov_info: ProvenanceDatasetInfo = prov.bipartite_xgraph.nodes[dataset_id]
        self.assertEqual(pred_info["dataset_type_name"], prov_info["dataset_type_name"])
        self.assertEqual(pred_info["data_id"], prov_info["data_id"])
        self.assertEqual(pred_info["run"], prov_info["run"])
        exists = prov_info["exists"]
        dataset_type_name = prov_info["dataset_type_name"]
        # We can remove this guard when we ingest QG-backed metadata and logs.
        if not dataset_type_name.endswith("_metadata") and not dataset_type_name.endswith("_log"):
            self.assertEqual(
                butler.get_dataset(dataset_id) is not None,
                exists,
                msg=(
                    f"Ingest/existence inconsistency for {dataset_type_name}"
                    f"@{prov_info['data_id']}/{dataset_id}]"
                ),
            )
        return exists

    def check_metadata(
        self, quantum_id: uuid.UUID, provenance_reader: ProvenanceQuantumGraphReader, butler: Butler
    ) -> None:
        """Check reading a metadata dataset from the provenance reader,
        and check that the original metadata file has been deleted.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            Unique ID for the quantum this metadata belongs to.
        provenance_reader : \
                `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraphReader`
            Reader for the provenance quantum graph.
        butler : `lsst.daf.butler.Butler`
            Client for the data repository.
        """
        # Try reading metadata through the quantum ID.
        (metadata1,) = provenance_reader.fetch_metadata([quantum_id]).values()
        self.assertIsInstance(metadata1, TaskMetadata)
        for _, dataset_id, pipeline_edges in provenance_reader.graph.bipartite_xgraph.out_edges(
            quantum_id, data="pipeline_edges"
        ):
            if pipeline_edges[0].connection_name == acc.METADATA_OUTPUT_CONNECTION_NAME:
                # Also try reading metadata through the dataset ID.
                (metadata2,) = provenance_reader.fetch_metadata([dataset_id]).values()
                break
        else:
            raise AssertionError("No metadata connection found.")
        self.assertEqual(metadata1, metadata2)

    def check_log(
        self, quantum_id: uuid.UUID, provenance_reader: ProvenanceQuantumGraphReader, butler: Butler
    ) -> None:
        """Check reading a log dataset from the provenance reader,
        and check that the original log file has been deleted.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            Unique ID for the quantum this log belongs to.
        provenance_reader : \
                `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraphReader`
            Reader for the provenance quantum graph.
        butler : `lsst.daf.butler.Butler`
            Client for the data repository.
        """
        # Try reading log through the quantum ID.
        (log1,) = provenance_reader.fetch_logs([quantum_id]).values()
        self.assertIsInstance(log1, ButlerLogRecords)
        for _, dataset_id, pipeline_edges in provenance_reader.graph.bipartite_xgraph.out_edges(
            quantum_id, data="pipeline_edges"
        ):
            if pipeline_edges[0].connection_name == acc.LOG_OUTPUT_CONNECTION_NAME:
                # Also try reading log through the dataset ID.
                (log2,) = provenance_reader.fetch_logs([dataset_id]).values()
                break
        else:
            raise AssertionError("No log connection found.")
        self.assertEqual(log1, log2)

    def check_packages(self, provenance_reader: ProvenanceQuantumGraphReader) -> None:
        """Check fetching package versions from the provenance graph.

        Parameters
        ----------
        provenance_reader : \
                `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraphReader`
            Reader for the provenance quantum graph.
        """
        packages = provenance_reader.fetch_packages()
        self.assertIsInstance(packages, Packages)
        self.assertIn("pipe_base", packages)

    def check_resource_usage_table(
        self, prov: ProvenanceQuantumGraph, expect_failure: bool, start_time: float
    ) -> None:
        """Check building a resource usage table from the provenance graph.

        Parameters
        ----------
        prov : `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraph`
            Reader for the provenance quantum graph.
        expect_failure : `bool`
            Whether to expect one quantum of 'calibrate' to fail (`True`) or
            succeed without writing anything (`False`).
        start_time : `float`
            A POSIX timestamp that strictly precedes the start time of any
            quantum's execution.
        """
        tbl = prov.make_task_resource_usage_table("calibrate", include_data_ids=True)
        if expect_failure:
            self.assertEqual(len(tbl), prov.header.n_task_quanta["calibrate"] - 1)
        else:
            self.assertEqual(len(tbl), prov.header.n_task_quanta["calibrate"])
        self.assertCountEqual(
            tbl.colnames,
            ["quantum_id"]
            + list(prov.pipeline_graph.tasks["calibrate"].dimensions.names)
            + list(QuantumResourceUsage.model_fields),
        )
        # Check that quantum start times are bounded by the before-execution
        # start_time and now.  This makes sure we didn't get any timezone
        # shenanigans.
        end_time = time.time()
        for quantum_start_time in tbl["start"]:
            self.assertGreater(quantum_start_time, start_time)
            self.assertLess(quantum_start_time, end_time)
        self.assertTrue(np.all(tbl["init_time"] >= 0.0))
        self.assertTrue(np.all(tbl["prep_time"] > 0.0))
        self.assertTrue(np.all(tbl["run_time"] >= 0.0))

    def check_quantum_table(self, prov: ProvenanceQuantumGraph, expect_failure: bool) -> None:
        """Check `ProvenanceQuantumGraph.make_quantum_table`.

        Parameters
        ----------
        prov : `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraph`
            Reader for the provenance quantum graph.
        expect_failure : `bool`
            Whether to expect one quantum of 'calibrate' to fail (`True`) or
            succeed without writing anything (`False`).
        """
        t = prov.make_quantum_table()
        self.assertTrue(np.all(t["Unknown"] == 0))
        self.assertEqual(list(t["Task"]), ["calibrate", "consolidate", "resample", "coadd"])
        self.assertEqual(t["TOTAL"][0], 8)
        self.assertEqual(t["EXPECTED"][0], 8)
        self.assertEqual(t["Blocked"][0], 0)
        self.assertEqual(t["TOTAL"][1], 2)
        self.assertEqual(t["EXPECTED"][1], 2)
        self.assertEqual(t["TOTAL"][2], 10)
        self.assertEqual(t["EXPECTED"][2], 10)
        if expect_failure:
            # calibrate
            self.assertEqual(t["Successful"][0], 7)
            self.assertEqual(t["Caveats"][0], "")
            self.assertEqual(t["Failed"][0], 1)
            # consolidate
            self.assertEqual(t["Successful"][1], 1)
            self.assertEqual(t["Caveats"][1], "")
            self.assertEqual(t["Failed"][1], 0)
            self.assertEqual(t["Blocked"][1], 1)
            # resample
            self.assertEqual(t["Successful"][2], 6)
            self.assertEqual(t["Caveats"][2], "")
            self.assertEqual(t["Failed"][2], 0)
            self.assertEqual(t["Blocked"][2], 4)
        else:
            # calibrate
            self.assertEqual(t["Successful"][0], 8)
            self.assertEqual(t["Caveats"][0], "*P(1)")
            self.assertEqual(t["Failed"][0], 0)
            # consolidate
            self.assertEqual(t["Successful"][1], 2)
            self.assertEqual(t["Caveats"][1], "")
            self.assertEqual(t["Failed"][1], 0)
            self.assertEqual(t["Blocked"][1], 0)
            # resample
            self.assertEqual(t["Successful"][2], 10)
            self.assertEqual(t["Caveats"][2], "*A(2)")
            self.assertEqual(t["Failed"][2], 0)
            self.assertEqual(t["Blocked"][2], 0)

    def check_exception_table(self, prov: ProvenanceQuantumGraph, expect_failure: bool) -> None:
        """Check `ProvenanceQuantumGraph.make_exception_table`.

        Parameters
        ----------
        prov : `lsst.pipe.base.quantum_graph.ProvenanceQuantumGraph`
            Reader for the provenance quantum graph.
        expect_failure : `bool`
            Whether to expect one quantum of 'calibrate' to fail (`True`) or
            succeed without writing anything (`False`).
        """
        t = prov.make_exception_table()
        if expect_failure:
            self.assertEqual(len(t), 0)
        else:
            self.assertEqual(list(t["Task"]), ["calibrate"])
            self.assertEqual(list(t["Exception"]), ["lsst.pipe.base.tests.mocks.MockAlgorithmError"])
            self.assertEqual(list(t["Count"]), [1])

    def test_all_successful(self) -> None:
        """Test running a full graph with no failures, and then scanning the
        results with assume_complete=False.

        Because there are no failures, assume_complete should not be necessary.
        """
        with self.make_test_repo() as prep:
            prep.config.assume_complete = False
            start_time = time.time()
            executed_quanta = list(
                self.iter_graph_execution(prep.butler_path, prep.predicted, raise_on_partial_outputs=False)
            )
            self.assertCountEqual(executed_quanta, prep.predicted.quantum_only_xgraph.nodes.keys())
            aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            with ProvenanceQuantumGraphReader.open(prep.config.output_path) as reader:
                self.check_provenance_graph(
                    prep.predicted, reader, prep.butler, expect_failure=False, start_time=start_time
                )

    def test_all_successful_two_phase(self) -> None:
        """Test running some of a graph with no failures, scanning with
        assume_complete=False, then finishing the graph and scanning again.
        """
        with self.make_test_repo() as prep:
            start_time = time.time()
            execution_iter = self.iter_graph_execution(
                prep.butler_path, prep.predicted, raise_on_partial_outputs=False
            )
            executed_quanta = list(itertools.islice(execution_iter, 9))
            self.assertEqual(len(executed_quanta), 9)
            # Run the scanner while telling it to assume failures might change,
            # so it just waits for incomplete quanta to finish (and then times
            # out).
            prep.config.assume_complete = False
            with self.assertRaises(RuntimeError):
                aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            # Finish executing the quanta.
            executed_quanta.extend(execution_iter)
            # Scan again, and write the provenance QG.
            aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            # Run the scanner again.
            with ProvenanceQuantumGraphReader.open(prep.config.output_path) as reader:
                self.check_provenance_graph(
                    prep.predicted, reader, prep.butler, expect_failure=False, start_time=start_time
                )

    def test_some_failed(self) -> None:
        """Test running a full graph with some failures, and then scanning the
        results with assume_complete=True.
        """
        with self.make_test_repo() as prep:
            prep.config.assume_complete = True
            start_time = time.time()
            for _ in self.iter_graph_execution(
                prep.butler_path, prep.predicted, raise_on_partial_outputs=True
            ):
                pass
            aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            with ProvenanceQuantumGraphReader.open(prep.config.output_path) as reader:
                self.check_provenance_graph(
                    prep.predicted, reader, prep.butler, expect_failure=True, start_time=start_time
                )

    def test_some_failed_two_phase(self) -> None:
        """Test running a full graph with some failures, then scanning the
        results with assume_complete=False, then scanning again with
        assume_complete=True.
        """
        with self.make_test_repo() as prep:
            start_time = time.time()
            for _ in self.iter_graph_execution(
                prep.butler_path, prep.predicted, raise_on_partial_outputs=True
            ):
                pass
            prep.config.assume_complete = False
            with self.assertRaisesRegex(RuntimeError, "1 quantum abandoned"):
                aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            prep.config.assume_complete = True
            aggregate_graph(prep.predicted_path, prep.butler_path, prep.config)
            with ProvenanceQuantumGraphReader.open(prep.config.output_path) as reader:
                self.check_provenance_graph(
                    prep.predicted, reader, prep.butler, expect_failure=True, start_time=start_time
                )

    def test_worker_failures(self) -> None:
        """Test that if failures occur on (multiple) workers we shut down
        gracefully instead of hanging.
        """
        with self.make_test_repo() as prep:
            with self.assertRaises(FatalWorkerError):
                aggregate_graph(prep.predicted_path, "nonexistent", prep.config)

    def test_cli_overrides(self) -> None:
        """Test that command-line options override config attributes as
        expected.
        """

        def mock_run(predicted_path: str, butler_path: str, config: AggregatorConfig) -> None:
            print(config.model_dump_json(indent=2))

        def check(result: Result, **kwargs: object) -> None:
            self.assertEqual(result.exit_code, 0, msg=result.output)
            self.assertEqual(
                result.output.strip(), AggregatorConfig(**kwargs).model_dump_json(indent=2).strip()
            )

        self.maxDiff = None
        runner = CliRunner()
        with unittest.mock.patch("lsst.pipe.base.quantum_graph.aggregator.aggregate_graph", mock_run):
            check(runner.invoke(aggregate_graph_cli, ("pg", "repo")))
            check(runner.invoke(aggregate_graph_cli, ("pg", "repo", "--output", "out")), output_path="out")
            check(runner.invoke(aggregate_graph_cli, ("pg", "repo", "-j", "4")), n_processes=4)
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--incomplete")),
                assume_complete=False,
            )
            check(runner.invoke(aggregate_graph_cli, ("pg", "repo", "--dry-run")), dry_run=True)
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--interactive-status")),
                interactive_status=True,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--log-status-interval", "120")),
                log_status_interval=120,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--no-register-dataset-types")),
                register_dataset_types=False,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--no-update-output-chain")),
                update_output_chain=False,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--worker-log-dir", "wlogs")),
                worker_log_dir="wlogs",
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--worker-log-level", "DEBUG")),
                worker_log_level="DEBUG",
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--zstd-level", "11")),
                zstd_level=11,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--zstd-dict-size", "143")),
                zstd_dict_size=143,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--zstd-dict-n-inputs", "2")),
                zstd_dict_n_inputs=2,
            )
            check(
                runner.invoke(aggregate_graph_cli, ("pg", "repo", "--mock-storage-classes")),
                mock_storage_classes=True,
            )


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Writer",)

import dataclasses
import enum
import itertools
import logging
import operator
import uuid
from typing import TypeVar

import networkx
import zstandard

from lsst.utils.packages import Packages

from ... import automatic_connection_constants as acc
from ...pipeline_graph import TaskImportMode
from .._common import BaseQuantumGraphWriter
from .._multiblock import Compressor, MultiblockWriter
from .._predicted import PredictedDatasetModel, PredictedQuantumGraphComponents, PredictedQuantumGraphReader
from .._provenance import (
    DATASET_ADDRESS_INDEX,
    DATASET_MB_NAME,
    LOG_ADDRESS_INDEX,
    LOG_MB_NAME,
    METADATA_ADDRESS_INDEX,
    METADATA_MB_NAME,
    QUANTUM_ADDRESS_INDEX,
    QUANTUM_MB_NAME,
    ProvenanceDatasetModel,
    ProvenanceInitQuantaModel,
    ProvenanceInitQuantumModel,
    ProvenanceQuantumModel,
)
from ._communicators import WriterCommunicator
from ._structs import ScanResult


class _CompressionState(enum.Enum):
    """Enumeration of the possible states of compression in `_ScanData`."""

    NOT_COMPRESSED = enum.auto()
    """Nothing is compressed."""

    LOG_AND_METADATA_COMPRESSED = enum.auto()
    """Only the logs and metadata are compressed."""

    ALL_COMPRESSED = enum.auto()
    """All `bytes` are compressed."""


@dataclasses.dataclass
class _ScanData:
    """Information from a quantum scan that has been partially processed for
    writing.
    """

    quantum_id: uuid.UUID
    """Unique ID of the quantum."""

    log_id: uuid.UUID
    """Unique ID of the log dataset."""

    metadata_id: uuid.UUID
    """Unique ID of the metadata dataset."""

    quantum: bytes = b""
    """Possibly-compressed JSON representation of the quantum provenance."""

    datasets: dict[uuid.UUID, bytes] = dataclasses.field(default_factory=dict)
    """Possibly-compressed JSON representation of output dataset provenance."""

    log: bytes = b""
    """Possibly-compressed log dataset content."""

    metadata: bytes = b""
    """Possibly-compressed metadata dataset content."""

    compression: _CompressionState = _CompressionState.NOT_COMPRESSED
    """Which data is compressed, if any."""

    def compress(self, compressor: Compressor) -> None:
        """Compress all data in place, if it isn't already.

        Parameters
        ----------
        compressor : `Compressor`
            Object that can compress `bytes`.
        """
        if self.compression is _CompressionState.NOT_COMPRESSED:
            self.metadata = compressor.compress(self.metadata)
            self.log = compressor.compress(self.log)
            self.compression = _CompressionState.LOG_AND_METADATA_COMPRESSED
        if self.compression is _CompressionState.LOG_AND_METADATA_COMPRESSED:
            self.quantum = compressor.compress(self.quantum)
            for key in self.datasets.keys():
                self.datasets[key] = compressor.compress(self.datasets[key])
        self.compression = _CompressionState.ALL_COMPRESSED


@dataclasses.dataclass
class _DataWriters:
    """A struct of low-level writer objects for the main components of a
    provenance quantum graph.

    Parameters
    ----------
    comms : `WriterCommunicator`
        Communicator helper object for the writer.
    predicted : `.PredictedQuantumGraphComponents`
        Components of the predicted graph.
    indices : `dict` [ `uuid.UUID`, `int` ]
        Mapping from UUID to internal integer ID, including both quanta and
        datasets.
    compressor : `Compressor`
        Object that can compress `bytes`.
    cdict_data : `bytes` or `None`, optional
        Bytes representation of the compression dictionary used by the
        compressor.
    """

    def __init__(
        self,
        comms: WriterCommunicator,
        predicted: PredictedQuantumGraphComponents,
        indices: dict[uuid.UUID, int],
        compressor: Compressor,
        cdict_data: bytes | None = None,
    ) -> None:
        assert comms.config.output_path is not None
        header = predicted.header.model_copy()
        header.graph_type = "provenance"
        self.graph = comms.enter(
            BaseQuantumGraphWriter.open(
                comms.config.output_path,
                header,
                predicted.pipeline_graph,
                indices,
                address_filename="nodes",
                compressor=compressor,
                cdict_data=cdict_data,
            ),
            on_close="Finishing writing provenance quantum graph.",
            is_progress_log=True,
        )
        self.graph.address_writer.addresses = [{}, {}, {}, {}]
        self.logs = comms.enter(
            MultiblockWriter.open_in_zip(self.graph.zf, LOG_MB_NAME, header.int_size, use_tempfile=True),
            on_close="Copying logs into zip archive.",
            is_progress_log=True,
        )
        self.graph.address_writer.addresses[LOG_ADDRESS_INDEX] = self.logs.addresses
        self.metadata = comms.enter(
            MultiblockWriter.open_in_zip(self.graph.zf, METADATA_MB_NAME, header.int_size, use_tempfile=True),
            on_close="Copying metadata into zip archive.",
            is_progress_log=True,
        )
        self.graph.address_writer.addresses[METADATA_ADDRESS_INDEX] = self.metadata.addresses
        self.datasets = comms.enter(
            MultiblockWriter.open_in_zip(self.graph.zf, DATASET_MB_NAME, header.int_size, use_tempfile=True),
            on_close="Copying dataset provenance into zip archive.",
            is_progress_log=True,
        )
        self.graph.address_writer.addresses[DATASET_ADDRESS_INDEX] = self.datasets.addresses
        self.quanta = comms.enter(
            MultiblockWriter.open_in_zip(self.graph.zf, QUANTUM_MB_NAME, header.int_size, use_tempfile=True),
            on_close="Copying quantum provenance into zip archive.",
            is_progress_log=True,
        )
        self.graph.address_writer.addresses[QUANTUM_ADDRESS_INDEX] = self.quanta.addresses

    graph: BaseQuantumGraphWriter
    """The parent graph writer."""

    datasets: MultiblockWriter
    """A writer for dataset provenance."""

    quanta: MultiblockWriter
    """A writer for quantum provenance."""

    metadata: MultiblockWriter
    """A writer for metadata content."""

    logs: MultiblockWriter
    """A writer for log content."""

    @property
    def compressor(self) -> Compressor:
        """Object that should be used to compress all JSON blocks."""
        return self.graph.compressor


@dataclasses.dataclass
class Writer:
    """A helper class for the provenance aggregator actually writes the
    provenance quantum graph file.
    """

    predicted_path: str
    """Path to the predicted quantum graph."""

    comms: WriterCommunicator
    """Communicator object for this worker."""

    predicted: PredictedQuantumGraphComponents = dataclasses.field(init=False)
    """Components of the predicted quantum graph."""

    existing_init_outputs: dict[uuid.UUID, set[uuid.UUID]] = dataclasses.field(default_factory=dict)
    """Mapping that tracks which init-outputs exist.

    This mapping is updated as scanners inform the writer about init-output
    existence, since we want to write that provenance information out only at
    the end.
    """

    indices: dict[uuid.UUID, int] = dataclasses.field(default_factory=dict)
    """Mapping from UUID to internal integer ID, including both quanta and
    datasets.

    This is fully initialized at construction.
    """

    output_dataset_ids: set[uuid.UUID] = dataclasses.field(default_factory=set)
    """The IDs of all datasets that are produced by this graph.

    This is fully initialized at construction.
    """

    overall_inputs: dict[uuid.UUID, PredictedDatasetModel] = dataclasses.field(default_factory=dict)
    """All datasets that are not produced by any quantum in this graph."""

    xgraph: networkx.DiGraph = dataclasses.field(default_factory=networkx.DiGraph)
    """A bipartite NetworkX graph linking datasets to quanta and quanta to
    datasets.

    This is fully initialized at construction.  There are no node or edge
    attributes in this graph; we only need it to store adjacency information
    with datasets as well as with quanta.
    """

    pending_compression_training: list[_ScanData] = dataclasses.field(default_factory=list)
    """Partially processed quantum scans that are being accumulated in order to
    build a compression dictionary.
    """

    def __post_init__(self) -> None:
        assert self.comms.config.output_path is not None, "Writer should not be used if writing is disabled."
        self.comms.log.info("Reading predicted quantum graph.")
        with PredictedQuantumGraphReader.open(
            self.predicted_path, import_mode=TaskImportMode.DO_NOT_IMPORT
        ) as reader:
            self.comms.check_for_cancel()
            reader.read_init_quanta()
            self.comms.check_for_cancel()
            reader.read_quantum_datasets()
            self.predicted = reader.components
        for predicted_init_quantum in self.predicted.init_quanta.root:
            self.existing_init_outputs[predicted_init_quantum.quantum_id] = set()
        self.comms.check_for_cancel()
        self.comms.log.info("Generating integer indexes and identifying outputs.")
        self._populate_indices_and_outputs()
        self.comms.check_for_cancel()
        self._populate_xgraph_and_inputs()
        self.comms.check_for_cancel()
        self.comms.log_progress(
            # We add one here for 'packages', which we do ingest but don't
            # record provenance for.
            logging.INFO,
            f"Graph has {len(self.output_dataset_ids) + 1} predicted output dataset(s).",
        )

    def _populate_indices_and_outputs(self) -> None:
        all_uuids = set(self.predicted.quantum_indices.keys())
        for quantum in self.comms.periodically_check_for_cancel(
            itertools.chain(
                self.predicted.init_quanta.root,
                self.predicted.quantum_datasets.values(),
            )
        ):
            if not quantum.task_label:
                # Skip the 'packages' producer quantum.
                continue
            all_uuids.update(quantum.iter_input_dataset_ids())
            self.output_dataset_ids.update(quantum.iter_output_dataset_ids())
        all_uuids.update(self.output_dataset_ids)
        self.indices = {
            node_id: node_index
            for node_index, node_id in self.comms.periodically_check_for_cancel(
                enumerate(sorted(all_uuids, key=operator.attrgetter("int")))
            )
        }

    def _populate_xgraph_and_inputs(self) -> None:
        for predicted_quantum in self.comms.periodically_check_for_cancel(
            itertools.chain(
                self.predicted.init_quanta.root,
                self.predicted.quantum_datasets.values(),
            )
        ):
            if not predicted_quantum.task_label:
                # Skip the 'packages' producer quantum.
                continue
            quantum_index = self.indices[predicted_quantum.quantum_id]
            for predicted_input in itertools.chain.from_iterable(predicted_quantum.inputs.values()):
                self.xgraph.add_edge(self.indices[predicted_input.dataset_id], quantum_index)
                if predicted_input.dataset_id not in self.output_dataset_ids:
                    self.overall_inputs.setdefault(predicted_input.dataset_id, predicted_input)
            for predicted_output in itertools.chain.from_iterable(predicted_quantum.outputs.values()):
                self.xgraph.add_edge(quantum_index, self.indices[predicted_output.dataset_id])

    @staticmethod
    def run(predicted_path: str, comms: WriterCommunicator) -> None:
        """Run the writer.

        Parameters
        ----------
        predicted_path : `str`
            Path to the predicted quantum graph.
        comms : `WriterCommunicator`
            Communicator for the writer.

        Notes
        -----
        This method is designed to run as the ``target`` in
        `WorkerContext.make_worker`.
        """
        with comms:
            writer = Writer(predicted_path, comms)
            writer.loop()

    def loop(self) -> None:
        """Run the main loop for the writer."""
        data_writers: _DataWriters | None = None
        if not self.comms.config.zstd_dict_size:
            data_writers = self.make_data_writers()
        self.comms.log.info("Polling for write requests from scanners.")
        for request in self.comms.poll():
            if data_writers is None:
                self.pending_compression_training.extend(self.make_scan_data(request))
                if len(self.pending_compression_training) >= self.comms.config.zstd_dict_n_inputs:
                    data_writers = self.make_data_writers()
            else:
                for scan_data in self.make_scan_data(request):
                    self.write_scan_data(scan_data, data_writers)
        if data_writers is None:
            data_writers = self.make_data_writers()
        self.write_init_outputs(data_writers)

    def make_data_writers(self) -> _DataWriters:
        """Make a compression dictionary, open the low-level writers, and
        write any accumulated scans that were needed to make the compression
        dictionary.

        Returns
        -------
        data_writers : `_DataWriters`
            Low-level writers struct.
        """
        cdict = self.make_compression_dictionary()
        self.comms.send_compression_dict(cdict.as_bytes())
        assert self.comms.config.output_path is not None
        self.comms.log.info("Opening output files.")
        data_writers = _DataWriters(
            self.comms,
            self.predicted,
            self.indices,
            compressor=zstandard.ZstdCompressor(self.comms.config.zstd_level, cdict),
            cdict_data=cdict.as_bytes(),
        )
        self.comms.check_for_cancel()
        self.comms.log.info("Compressing and writing queued scan requests.")
        for scan_data in self.pending_compression_training:
            self.write_scan_data(scan_data, data_writers)
        del self.pending_compression_training
        self.comms.check_for_cancel()
        self.write_overall_inputs(data_writers)
        self.write_packages(data_writers)
        self.comms.log.info("Returning to write request loop.")
        return data_writers

    def make_compression_dictionary(self) -> zstandard.ZstdCompressionDict:
        """Make the compression dictionary.

        Returns
        -------
        cdict : `zstandard.ZstdCompressionDict`
            The compression dictionary.
        """
        if (
            not self.comms.config.zstd_dict_size
            or len(self.pending_compression_training) < self.comms.config.zstd_dict_n_inputs
        ):
            self.comms.log.info("Making compressor with no dictionary.")
            return zstandard.ZstdCompressionDict(b"")
        self.comms.log.info("Training compression dictionary.")
        training_inputs: list[bytes] = []
        # We start the dictionary training with *predicted* quantum dataset
        # models, since those have almost all of the same attributes as the
        # provenance quantum and dataset models, and we can get a nice random
        # sample from just the first N, since they're ordered by UUID.  We
        # chop out the datastore records since those don't appear in the
        # provenance graph.
        for predicted_quantum in self.predicted.quantum_datasets.values():
            if len(training_inputs) == self.comms.config.zstd_dict_n_inputs:
                break
            predicted_quantum.datastore_records.clear()
            training_inputs.append(predicted_quantum.model_dump_json().encode())
        # Add the provenance quanta, metadata, and logs we've accumulated.
        for scan_data in self.pending_compression_training:
            assert scan_data.compression is _CompressionState.NOT_COMPRESSED
            training_inputs.append(scan_data.quantum)
            training_inputs.append(scan_data.metadata)
            training_inputs.append(scan_data.log)
        return zstandard.train_dictionary(self.comms.config.zstd_dict_size, training_inputs)

    def write_init_outputs(self, data_writers: _DataWriters) -> None:
        """Write provenance for init-output datasets and init-quanta.

        Parameters
        ----------
        data_writers : `_DataWriters`
            Low-level writers struct.
        """
        self.comms.log.info("Writing init outputs.")
        init_quanta = ProvenanceInitQuantaModel()
        for predicted_init_quantum in self.predicted.init_quanta.root:
            if not predicted_init_quantum.task_label:
                # Skip the 'packages' producer quantum.
                continue
            existing_outputs = self.existing_init_outputs[predicted_init_quantum.quantum_id]
            for predicted_output in itertools.chain.from_iterable(predicted_init_quantum.outputs.values()):
                dataset_index = self.indices[predicted_output.dataset_id]
                provenance_output = ProvenanceDatasetModel.from_predicted(
                    predicted_output,
                    producer=self.indices[predicted_init_quantum.quantum_id],
                    consumers=self.xgraph.successors(dataset_index),
                )
                provenance_output.exists = predicted_output.dataset_id in existing_outputs
                data_writers.datasets.write_model(
                    provenance_output.dataset_id, provenance_output, data_writers.compressor
                )
            init_quanta.root.append(
                ProvenanceInitQuantumModel.from_predicted(predicted_init_quantum, self.indices)
            )
        data_writers.graph.write_single_model("init_quanta", init_quanta)

    def write_overall_inputs(self, data_writers: _DataWriters) -> None:
        """Write provenance for overall-input datasets.

        Parameters
        ----------
        data_writers : `_DataWriters`
            Low-level writers struct.
        """
        self.comms.log.info("Writing overall inputs.")
        for predicted_input in self.comms.periodically_check_for_cancel(self.overall_inputs.values()):
            if predicted_input.dataset_id not in data_writers.datasets.addresses:
                dataset_index = self.indices[predicted_input.dataset_id]
                data_writers.datasets.write_model(
                    predicted_input.dataset_id,
                    ProvenanceDatasetModel.from_predicted(
                        predicted_input,
                        producer=None,
                        consumers=self.xgraph.successors(dataset_index),
                    ),
                    data_writers.compressor,
                )
        del self.overall_inputs

    @staticmethod
    def write_packages(data_writers: _DataWriters) -> None:
        """Write package version information to the provenance graph.

        Parameters
        ----------
        data_writers : `_DataWriters`
            Low-level writers struct.
        """
        packages = Packages.fromSystem(include_all=True)
        data = packages.toBytes("json")
        data_writers.graph.write_single_block("packages", data)

    def make_scan_data(self, request: ScanResult) -> list[_ScanData]:
        """Process a `ScanResult` into `_ScanData`.

        Parameters
        ----------
        request : `ScanResult`
            Result of a quantum scan.

        Returns
        -------
        data : `list` [ `_ScanData` ]
            A zero- or single-element list of `_ScanData` to write or save for
            compression-dict training.  A zero-element list is returned if the
            scan actually represents an init quantum.
        """
        if (existing_init_outputs := self.existing_init_outputs.get(request.quantum_id)) is not None:
            self.comms.log.debug("Handling init-output scan for %s.", request.quantum_id)
            existing_init_outputs.update(request.existing_outputs)
            self.comms.report_write()
            return []
        self.comms.log.debug("Handling quantum scan for %s.", request.quantum_id)
        predicted_quantum = self.predicted.quantum_datasets[request.quantum_id]
        quantum_index = self.indices[predicted_quantum.quantum_id]
        (metadata_output,) = predicted_quantum.outputs[acc.METADATA_OUTPUT_CONNECTION_NAME]
        (log_output,) = predicted_quantum.outputs[acc.LOG_OUTPUT_CONNECTION_NAME]
        data = _ScanData(
            request.quantum_id,
            metadata_id=metadata_output.dataset_id,
            log_id=log_output.dataset_id,
            compression=(
                _CompressionState.LOG_AND_METADATA_COMPRESSED
                if request.is_compressed
                else _CompressionState.NOT_COMPRESSED
            ),
        )
        for predicted_output in itertools.chain.from_iterable(predicted_quantum.outputs.values()):
            dataset_index = self.indices[predicted_output.dataset_id]
            provenance_output = ProvenanceDatasetModel.from_predicted(
                predicted_output,
                producer=quantum_index,
                consumers=self.xgraph.successors(dataset_index),
            )
            provenance_output.exists = provenance_output.dataset_id in request.existing_outputs
            data.datasets[provenance_output.dataset_id] = provenance_output.model_dump_json().encode()
        provenance_quantum = ProvenanceQuantumModel.from_predicted(predicted_quantum, self.indices)
        provenance_quantum.status = request.get_run_status()
        provenance_quantum.caveats = request.caveats
        provenance_quantum.exception = request.exception
        provenance_quantum.resource_usage = request.resource_usage
        data.quantum = provenance_quantum.model_dump_json().encode()
        data.metadata = request.metadata
        data.log = request.log
        return [data]

    def write_scan_data(self, scan_data: _ScanData, data_writers: _DataWriters) -> None:
        """Write scan data to the provenance graph.

        Parameters
        ----------
        scan_data : `_ScanData`
            Preprocessed information to write.
        data_writers : `_DataWriters`
            Low-level writers struct.
        """
        self.comms.log.debug("Writing quantum %s.", scan_data.quantum_id)
        scan_data.compress(data_writers.compressor)
        data_writers.quanta.write_bytes(scan_data.quantum_id, scan_data.quantum)
        for dataset_id, dataset_data in scan_data.datasets.items():
            data_writers.datasets.write_bytes(dataset_id, dataset_data)
        if scan_data.metadata:
            address = data_writers.metadata.write_bytes(scan_data.quantum_id, scan_data.metadata)
            data_writers.metadata.addresses[scan_data.metadata_id] = address
        if scan_data.log:
            address = data_writers.logs.write_bytes(scan_data.quantum_id, scan_data.log)
            data_writers.logs.addresses[scan_data.log_id] = address
        # We shouldn't need this predicted quantum anymore; delete it in the
        # hopes that'll free up some memory.
        del self.predicted.quantum_datasets[scan_data.quantum_id]
        self.comms.report_write()


_T = TypeVar("_T")



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Ingester",)

import dataclasses
import logging
import time
import uuid
from collections import defaultdict

from lsst.daf.butler import Butler, CollectionType, DatasetRef, DimensionGroup
from lsst.daf.butler.datastore.record_data import DatastoreRecordData
from lsst.daf.butler.registry import ConflictingDefinitionError

from ...pipeline_graph import TaskImportMode
from .._common import DatastoreName
from .._predicted import PredictedDatasetModel, PredictedQuantumGraphComponents, PredictedQuantumGraphReader
from ._communicators import IngesterCommunicator


@dataclasses.dataclass
class Ingester:
    """A helper class for the provenance aggregator that handles ingestion into
    the central butler repository.
    """

    predicted_path: str
    """Path to the predicted quantum graph."""

    butler_path: str
    """Path or alias to the central butler repository."""

    comms: IngesterCommunicator
    """Communicator object for this worker."""

    predicted: PredictedQuantumGraphComponents = dataclasses.field(init=False)
    """Components of the predicted graph."""

    butler: Butler = dataclasses.field(init=False)
    """Client for the central butler repository."""

    n_datasets_ingested: int = 0
    """Total number of datasets ingested by this invocation."""

    n_datasets_skipped: int = 0
    """Total number of datasets skipped because they were already present."""

    n_producers_pending: int = 0
    """Number of quanta whose outputs are currently pending ingest."""

    refs_pending: defaultdict[DimensionGroup, list[DatasetRef]] = dataclasses.field(
        default_factory=lambda: defaultdict(list)
    )
    """Dataset references pending ingest, grouped by their dimensions."""

    records_pending: dict[DatastoreName, DatastoreRecordData] = dataclasses.field(default_factory=dict)
    """Datastore records pending ingest, grouped by datastore name."""

    already_ingested: set[uuid.UUID] | None = None
    """A set of all dataset IDs already present in the output RUN
    collection.

    If this is not `None`, the ingester is in defensive ingest mode, either
    because it was configured to query for these dataset IDs up front, or
    because a transaction failed due to a dataset already being present.
    """

    last_ingest_time: float = dataclasses.field(default_factory=time.time)
    """POSIX timestamp since the last ingest transaction concluded."""

    def __post_init__(self) -> None:
        self.comms.log.verbose("Reading from predicted quantum graph.")
        with PredictedQuantumGraphReader.open(
            self.predicted_path, import_mode=TaskImportMode.DO_NOT_IMPORT
        ) as reader:
            # We only need the header and pipeline graph.
            self.predicted = reader.components
        if self.comms.config.mock_storage_classes:
            import lsst.pipe.base.tests.mocks  # noqa: F401
        self.comms.log.verbose("Initializing butler.")
        self.butler = Butler.from_config(self.butler_path, writeable=not self.comms.config.dry_run)

    @property
    def n_datasets_pending(self) -> int:
        """The number of butler datasets currently pending."""
        return sum(len(v) for v in self.refs_pending.values())

    @staticmethod
    def run(predicted_path: str, butler_path: str, comms: IngesterCommunicator) -> None:
        """Run the ingester.

        Parameters
        ----------
        predicted_path : `str`
            Path to the predicted quantum graph.
        butler_path : `str`
            Path or alias to the central butler repository.
        comms : `IngesterCommunicator`
            Communicator for the ingester.

        Notes
        -----
        This method is designed to run as the ``target`` in
        `WorkerContext.make_worker`.
        """
        with comms:
            ingester = Ingester(predicted_path, butler_path, comms)
            ingester.loop()

    def loop(self) -> None:
        """Run the main loop for the ingester."""
        self.comms.log.verbose("Registering collections and dataset types.")
        if not self.comms.config.dry_run:
            if self.comms.config.register_dataset_types:
                self.predicted.pipeline_graph.register_dataset_types(
                    self.butler,
                    include_inputs=False,
                    include_packages=True,
                    include_configs=True,
                    include_logs=True,
                )
            self.butler.collections.register(self.predicted.header.output_run)
            # Updating the output chain cannot happen inside the caching
            # context.
            if self.comms.config.update_output_chain:
                self.update_output_chain()
        with self.butler.registry.caching_context():
            if self.comms.config.defensive_ingest:
                self.fetch_already_ingested()
            self.comms.log.info("Startup completed in %ss.", time.time() - self.last_ingest_time)
            self.last_ingest_time = time.time()
            for ingest_request in self.comms.poll():
                self.n_producers_pending += 1
                self.comms.log.debug(f"Got ingest request for producer {ingest_request.producer_id}.")
                self.update_pending(ingest_request.datasets, ingest_request.records)
                if self.n_datasets_pending > self.comms.config.ingest_batch_size:
                    self.ingest()
            self.comms.log.info("All ingest requests received.")
            # We use 'while' in case this fails with a conflict and we switch
            # to defensive mode (should be at most two iterations).
            ingest_start_time = time.time()
            while self.n_datasets_pending:
                n_datasets = self.n_datasets_pending
                self.ingest()
                self.comms.log.verbose(
                    "Gathered %d final datasets in %ss and ingested them in %ss.",
                    n_datasets,
                    ingest_start_time - self.last_ingest_time,
                    time.time() - ingest_start_time,
                )
            if self.n_producers_pending:
                # We can finish with returns pending if we filtered out all of
                # the datasets we started with as already existing.
                self.report()
        self.comms.log_progress(
            logging.INFO,
            f"Ingested {self.n_datasets_ingested} dataset(s); "
            f"skipped {self.n_datasets_skipped} already present.",
        )

    def ingest(self) -> None:
        """Ingest all pending datasets and report success to the supervisor."""
        ingest_start_time = time.time()
        self.comms.log.verbose(
            "Gathered %d datasets from %d quanta in %ss.",
            self.n_datasets_pending,
            self.n_producers_pending,
            ingest_start_time - self.last_ingest_time,
        )
        try:
            if not self.comms.config.dry_run:
                with self.butler.registry.transaction():
                    for refs in self.refs_pending.values():
                        self.butler.registry._importDatasets(refs, expand=False, assume_new=True)
                    self.butler._datastore.import_records(self.records_pending)
            self.last_ingest_time = time.time()
            self.comms.log.verbose(
                "Ingested %d datasets from %d quanta in %ss.",
                self.n_datasets_pending,
                self.n_producers_pending,
                self.last_ingest_time - ingest_start_time,
            )
            self.n_datasets_ingested += self.n_datasets_pending
        except ConflictingDefinitionError:
            if self.already_ingested is None:
                self.comms.log_progress(
                    logging.INFO,
                    "Some outputs seem to have already been ingested; querying for existing datasets and "
                    "switching to defensive ingest mode.",
                )
                self.fetch_already_ingested()
                # We just return instead of trying again immediately because we
                # might have just shrunk the number of pending datasets below
                # the batch threshold.
                return
            else:
                raise
        self.report()
        self.records_pending.clear()
        self.refs_pending.clear()

    def report(self) -> None:
        """Report a successful ingest to the supervisor."""
        self.comms.report_ingest(self.n_producers_pending)
        self.n_producers_pending = 0

    def fetch_already_ingested(self) -> None:
        """Query for the UUIDs of all dataset already present in the output
        RUN collection, and filter and pending datasets accordingly.
        """
        self.comms.log.info("Fetching all UUIDs in output collection %r.", self.predicted.header.output_run)
        self.already_ingested = set(
            self.butler.registry._fetch_run_dataset_ids(self.predicted.header.output_run)
        )
        kept: set[uuid.UUID] = set()
        for dimensions, refs in self.refs_pending.items():
            filtered_refs: list[DatasetRef] = []
            for ref in refs:
                if ref.id not in self.already_ingested:
                    kept.add(ref.id)
                    filtered_refs.append(ref)
                else:
                    self.n_datasets_skipped += 1
            self.refs_pending[dimensions] = filtered_refs
        for datastore_name, datastore_records in list(self.records_pending.items()):
            if (filtered_records := datastore_records.subset(kept)) is not None:
                self.records_pending[datastore_name] = filtered_records
            else:
                del self.records_pending[datastore_name]

    def update_pending(
        self, datasets: list[PredictedDatasetModel], records: dict[DatastoreName, DatastoreRecordData]
    ) -> None:
        """Add an ingest request to the pending-ingest data structures.

        Parameters
        ----------
        datasets : `list` [ `PredictedDatasetModel` ]
            Registry information about the datasets.
        records : `dict` [ `str`, \
                `lsst.daf.butler.datastore.record_data.DatastoreRecordData` ]
            Datastore information about the datasets.
        """
        n_given = len(datasets)
        if self.already_ingested is not None:
            datasets = [d for d in datasets if d.dataset_id not in self.already_ingested]
            kept = {d.dataset_id for d in datasets}
            self.n_datasets_skipped += n_given - len(kept)
            records = {
                datastore_name: filtered_records
                for datastore_name, original_records in records.items()
                if (filtered_records := original_records.subset(kept)) is not None
            }
        for dataset in datasets:
            ref = self.predicted.make_dataset_ref(dataset)
            self.refs_pending[ref.datasetType.dimensions].append(ref)
        for datastore_name, datastore_records in records.items():
            if (existing_records := self.records_pending.get(datastore_name)) is not None:
                existing_records.update(datastore_records)
            else:
                self.records_pending[datastore_name] = datastore_records

    def update_output_chain(self) -> None:
        """Update the output CHAINED collection to include the output RUN
        collection (and the inputs, if the output CHAINED collection does not
        exist).

        Notes
        -----
        This method cannot be called inside the registry caching context.
        """
        if self.predicted.header.output is None:
            return
        self.comms.log.info(
            "Updating output collection %s to include %s.",
            self.predicted.header.output,
            self.predicted.header.output_run,
        )
        if self.butler.collections.register(self.predicted.header.output, CollectionType.CHAINED):
            # Chain is new; need to add inputs, but we want to flatten them
            # first.
            if self.predicted.header.inputs:
                flattened = self.butler.collections.query(self.predicted.header.inputs, flatten_chains=True)
                self.butler.collections.extend_chain(self.predicted.header.output, flattened)
        self.butler.collections.prepend_chain(self.predicted.header.output, self.predicted.header.output_run)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "ProvenanceDatasetInfo",
    "ProvenanceDatasetModel",
    "ProvenanceInitQuantumInfo",
    "ProvenanceInitQuantumModel",
    "ProvenanceQuantumGraph",
    "ProvenanceQuantumGraphReader",
    "ProvenanceQuantumInfo",
    "ProvenanceQuantumModel",
)


import dataclasses
import sys
import uuid
from collections import Counter
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Self, TypedDict

import astropy.table
import networkx
import numpy as np
import pydantic

from lsst.daf.butler import DataCoordinate
from lsst.resources import ResourcePathExpression
from lsst.utils.packages import Packages

from .._status import QuantumSuccessCaveats
from ..pipeline_graph import PipelineGraph, TaskImportMode, TaskInitNode
from ..quantum_provenance_graph import ExceptionInfo, QuantumRunStatus
from ..resource_usage import QuantumResourceUsage
from ._common import (
    BaseQuantumGraph,
    BaseQuantumGraphReader,
    ConnectionName,
    DataCoordinateValues,
    DatasetIndex,
    DatasetInfo,
    DatasetTypeName,
    HeaderModel,
    QuantumIndex,
    QuantumInfo,
    TaskLabel,
)
from ._multiblock import AddressReader, MultiblockReader
from ._predicted import PredictedDatasetModel, PredictedQuantumDatasetsModel

if TYPE_CHECKING:
    from lsst.daf.butler.logging import ButlerLogRecords

    from .._task_metadata import TaskMetadata


DATASET_ADDRESS_INDEX = 0
QUANTUM_ADDRESS_INDEX = 1
LOG_ADDRESS_INDEX = 2
METADATA_ADDRESS_INDEX = 3

DATASET_MB_NAME = "datasets"
QUANTUM_MB_NAME = "quanta"
LOG_MB_NAME = "logs"
METADATA_MB_NAME = "metadata"


class ProvenanceDatasetInfo(DatasetInfo):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    node data for a provenance dataset.

    Since NetworkX types are not generic over their node mapping type, this has
    to be used explicitly, e.g.::

        node_data: ProvenanceDatasetInfo = xgraph.nodes[dataset_id]

    where ``xgraph`` is `ProvenanceQuantumGraph.bipartite_xgraph`.
    """

    dataset_id: uuid.UUID
    """Unique identifier for the dataset."""

    exists: bool
    """Whether this dataset existed immediately after the quantum graph was
    run.

    This is always `True` for overall input datasets.  It is also `True` for
    datasets that were produced and then removed before/during transfer back to
    the central butler repository.
    """


class ProvenanceQuantumInfo(QuantumInfo):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    node data for a provenance quantum.

    Since NetworkX types are not generic over their node mapping type, this has
    to be used explicitly, e.g.::

        node_data: ProvenanceQuantumInfo = xgraph.nodes[quantum_id]

    where ``xgraph`` is `ProvenanceQuantumGraph.bipartite_xgraph` or
    `ProvenanceQuantumGraph.quantum_only_xgraph`
    """

    status: QuantumRunStatus
    """Enumerated status for the quantum."""

    caveats: QuantumSuccessCaveats | None
    """Flags indicating caveats on successful quanta."""

    exception: ExceptionInfo | None
    """Information about an exception raised when the quantum was executing."""

    resource_usage: QuantumResourceUsage | None
    """Resource usage information (timing, memory use) for this quantum."""


class ProvenanceInitQuantumInfo(TypedDict):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    node data for a provenance init quantum.

    Since NetworkX types are not generic over their node mapping type, this has
    to be used explicitly, e.g.::

        node_data: ProvenanceInitQuantumInfo = xgraph.nodes[quantum_id]

    where ``xgraph`` is `ProvenanceQuantumGraph.bipartite_xgraph`.
    """

    data_id: DataCoordinate
    """Data ID of the quantum.

    This is always an empty ID; this key exists to allow init-quanta and
    regular quanta to be treated more similarly.
    """

    task_label: str
    """Label of the task for this quantum."""

    pipeline_node: TaskInitNode
    """Node in the pipeline graph for this task's init-only step."""


class ProvenanceDatasetModel(PredictedDatasetModel):
    """Data model for the datasets in a provenance quantum graph file."""

    exists: bool
    """Whether this dataset existed immediately after the quantum graph was
    run.

    This is always `True` for overall input datasets.  It is also `True` for
    datasets that were produced and then removed before/during transfer back to
    the central butler repository.
    """

    producer: QuantumIndex | None = None
    """Internal integer ID of the quantum that produced this dataset.

    This is `None` for overall inputs to the graph.
    """

    consumers: list[QuantumIndex] = pydantic.Field(default_factory=list)
    """Internal integer IDs of quanta that were predicted to consume this
    dataset.
    """

    @property
    def node_id(self) -> uuid.UUID:
        """Alias for the dataset ID."""
        return self.dataset_id

    @classmethod
    def from_predicted(
        cls,
        predicted: PredictedDatasetModel,
        producer: QuantumIndex | None = None,
        consumers: Iterable[QuantumIndex] = (),
    ) -> ProvenanceDatasetModel:
        """Construct from a predicted dataset model.

        Parameters
        ----------
        predicted : `PredictedDatasetModel`
            Information about the dataset from the predicted graph.
        producer : `int` or `None`, optional
            Internal ID of the quantum that was predicted to produce this
            dataset.
        consumers : `~collections.abc.Iterable` [`int`], optional
            Internal IDs of the quanta that were predicted to consume this
            dataset.

        Returns
        -------
        provenance : `ProvenanceDatasetModel`
            Provenance dataset model.

        Notes
        -----
        This initializes `exists` to `True` when ``producer is None`` and
        `False` otherwise, on the assumption that it will be updated later.
        """
        return cls.model_construct(
            dataset_id=predicted.dataset_id,
            dataset_type_name=predicted.dataset_type_name,
            data_coordinate=predicted.data_coordinate,
            run=predicted.run,
            exists=(producer is None),  # if it's not produced by this QG, it's an overall input
            producer=producer,
            consumers=list(consumers),
        )

    def _add_to_graph(self, graph: ProvenanceQuantumGraph, address_reader: AddressReader) -> None:
        """Add this dataset and its edges to quanta to a provenance graph.

        Parameters
        ----------
        graph : `ProvenanceQuantumGraph`
            Graph to update in place.
        address_reader : `AddressReader`
            Reader object that can be used to look up UUIDs from integer
            indexes.

        Notes
        -----
        This method adds:

        - a ``bipartite_xgraph`` dataset node with full attributes;
        - ``bipartite_xgraph`` edges to adjacent quanta (which adds quantum
          nodes with no attributes), without populating edge attributes;
        - ``quantum_only_xgraph`` edges for each pair of quanta in which one
          produces this dataset and another consumes it (this also adds quantum
          nodes with no attributes).
        """
        dataset_type_node = graph.pipeline_graph.dataset_types[self.dataset_type_name]
        data_id = DataCoordinate.from_full_values(dataset_type_node.dimensions, tuple(self.data_coordinate))
        graph._bipartite_xgraph.add_node(
            self.dataset_id,
            data_id=data_id,
            dataset_type_name=self.dataset_type_name,
            pipeline_node=dataset_type_node,
            run=self.run,
            exists=self.exists,
        )
        producer_id: uuid.UUID | None = None
        if self.producer is not None:
            producer_id = address_reader.find(self.producer).key
            graph._bipartite_xgraph.add_edge(producer_id, self.dataset_id)
        for consumer_index in self.consumers:
            consumer_id = address_reader.find(consumer_index).key
            graph._bipartite_xgraph.add_edge(self.dataset_id, consumer_id)
            if producer_id is not None:
                graph._quantum_only_xgraph.add_edge(producer_id, consumer_id)
        graph._datasets_by_type[self.dataset_type_name][data_id] = self.dataset_id

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class ProvenanceQuantumModel(pydantic.BaseModel):
    """Data model for the quanta in a provenance quantum graph file."""

    quantum_id: uuid.UUID
    """Unique identifier for the quantum."""

    task_label: TaskLabel
    """Name of the type of this dataset.

    This is always a parent dataset type name, not a component.

    Note that full dataset type definitions are stored in the pipeline graph.
    """

    data_coordinate: DataCoordinateValues = pydantic.Field(default_factory=list)
    """The full values (required and implied) of this dataset's data ID."""

    status: QuantumRunStatus = QuantumRunStatus.METADATA_MISSING
    """Enumerated status for the quantum."""

    caveats: QuantumSuccessCaveats | None = None
    """Flags indicating caveats on successful quanta."""

    exception: ExceptionInfo | None = None
    """Information about an exception raised when the quantum was executing."""

    inputs: dict[ConnectionName, list[DatasetIndex]] = pydantic.Field(default_factory=dict)
    """Internal integer IDs of the datasets predicted to be consumed by this
    quantum, grouped by connection name.
    """

    outputs: dict[ConnectionName, list[DatasetIndex]] = pydantic.Field(default_factory=dict)
    """Internal integer IDs of the datasets predicted to be produced by this
    quantum, grouped by connection name.
    """

    resource_usage: QuantumResourceUsage | None = None
    """Resource usage information (timing, memory use) for this quantum."""

    @property
    def node_id(self) -> uuid.UUID:
        """Alias for the quantum ID."""
        return self.quantum_id

    @classmethod
    def from_predicted(
        cls, predicted: PredictedQuantumDatasetsModel, indices: Mapping[uuid.UUID, int]
    ) -> ProvenanceQuantumModel:
        """Construct from a predicted quantum model.

        Parameters
        ----------
        predicted : `PredictedQuantumDatasetsModel`
            Information about the quantum from the predicted graph.
        indices : `~collections.abc.Mapping [`uuid.UUID`, `int`]
            Mapping from quantum or dataset UUID to internal integer ID.

        Returns
        -------
        provenance : `ProvenanceQuantumModel`
            Provenance quantum model.
        """
        inputs = {
            connection_name: [indices[d.dataset_id] for d in predicted_inputs]
            for connection_name, predicted_inputs in predicted.inputs.items()
        }
        outputs = {
            connection_name: [indices[d.dataset_id] for d in predicted_outputs]
            for connection_name, predicted_outputs in predicted.outputs.items()
        }
        return cls(
            quantum_id=predicted.quantum_id,
            task_label=predicted.task_label,
            data_coordinate=predicted.data_coordinate,
            inputs=inputs,
            outputs=outputs,
        )

    def _add_to_graph(self, graph: ProvenanceQuantumGraph, address_reader: AddressReader) -> None:
        """Add this quantum and its edges to datasets to a provenance graph.

        Parameters
        ----------
        graph : `ProvenanceQuantumGraph`
            Graph to update in place.
        address_reader : `AddressReader`
            Reader object that can be used to look up UUIDs from integer
            indexes.

        Notes
        -----
        This method adds:

        - a ``bipartite_xgraph`` quantum node with full attributes;
        - a ``quantum_only_xgraph`` quantum node with full attributes;
        - ``bipartite_xgraph`` edges to adjacent datasets (which adds datasets
          nodes with no attributes), while populating those edge attributes;
        - ``quantum_only_xgraph`` edges to any adjacent quantum that has also
          already been loaded.
        """
        task_node = graph.pipeline_graph.tasks[self.task_label]
        data_id = DataCoordinate.from_full_values(task_node.dimensions, tuple(self.data_coordinate))
        graph._bipartite_xgraph.add_node(
            self.quantum_id,
            data_id=data_id,
            task_label=self.task_label,
            pipeline_node=task_node,
            status=self.status,
            caveats=self.caveats,
            exception=self.exception,
            resource_usage=self.resource_usage,
        )
        for connection_name, dataset_indices in self.inputs.items():
            read_edge = task_node.get_input_edge(connection_name)
            for dataset_index in dataset_indices:
                dataset_id = address_reader.find(dataset_index).key
                graph._bipartite_xgraph.add_edge(dataset_id, self.quantum_id, is_read=True)
                graph._bipartite_xgraph.edges[dataset_id, self.quantum_id].setdefault(
                    "pipeline_edges", []
                ).append(read_edge)
        for connection_name, dataset_indices in self.outputs.items():
            write_edge = task_node.get_output_edge(connection_name)
            for dataset_index in dataset_indices:
                dataset_id = address_reader.find(dataset_index).key
                graph._bipartite_xgraph.add_edge(
                    self.quantum_id,
                    dataset_id,
                    is_read=False,
                    # There can only be one pipeline edge for an output.
                    pipeline_edges=[write_edge],
                )
        graph._quanta_by_task_label[self.task_label][data_id] = self.quantum_id
        graph._quantum_only_xgraph.add_node(self.quantum_id, **graph._bipartite_xgraph.nodes[self.quantum_id])
        for dataset_id in graph._bipartite_xgraph.predecessors(self.quantum_id):
            for upstream_quantum_id in graph._bipartite_xgraph.predecessors(dataset_id):
                graph._quantum_only_xgraph.add_edge(upstream_quantum_id, self.quantum_id)
        for dataset_id in graph._bipartite_xgraph.successors(self.quantum_id):
            for downstream_quantum_id in graph._bipartite_xgraph.successors(dataset_id):
                graph._quantum_only_xgraph.add_edge(self.quantum_id, downstream_quantum_id)

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class ProvenanceInitQuantumModel(pydantic.BaseModel):
    """Data model for the special "init" quanta in a provenance quantum graph
    file.
    """

    quantum_id: uuid.UUID
    """Unique identifier for the quantum."""

    task_label: TaskLabel
    """Name of the type of this dataset.

    This is always a parent dataset type name, not a component.

    Note that full dataset type definitions are stored in the pipeline graph.
    """

    inputs: dict[ConnectionName, DatasetIndex] = pydantic.Field(default_factory=dict)
    """Internal integer IDs of the datasets predicted to be consumed by this
    quantum, grouped by connection name.
    """

    outputs: dict[ConnectionName, DatasetIndex] = pydantic.Field(default_factory=dict)
    """Internal integer IDs of the datasets predicted to be produced by this
    quantum, grouped by connection name.
    """

    @classmethod
    def from_predicted(
        cls, predicted: PredictedQuantumDatasetsModel, indices: Mapping[uuid.UUID, int]
    ) -> ProvenanceInitQuantumModel:
        """Construct from a predicted quantum model.

        Parameters
        ----------
        predicted : `PredictedQuantumDatasetsModel`
            Information about the quantum from the predicted graph.
        indices : `~collections.abc.Mapping [`uuid.UUID`, `int`]
            Mapping from quantum or dataset UUID to internal integer ID.

        Returns
        -------
        provenance : `ProvenanceInitQuantumModel`
            Provenance init quantum model.
        """
        inputs = {
            connection_name: indices[predicted_inputs[0].dataset_id]
            for connection_name, predicted_inputs in predicted.inputs.items()
        }
        outputs = {
            connection_name: indices[predicted_outputs[0].dataset_id]
            for connection_name, predicted_outputs in predicted.outputs.items()
        }
        return cls(
            quantum_id=predicted.quantum_id,
            task_label=predicted.task_label,
            inputs=inputs,
            outputs=outputs,
        )

    def _add_to_graph(
        self,
        graph: ProvenanceQuantumGraph,
        address_reader: AddressReader,
        empty_data_id: DataCoordinate,
    ) -> None:
        """Add this quantum and its edges to datasets to a provenance graph.

        Parameters
        ----------
        graph : `ProvenanceQuantumGraph`
            Graph to update in place.
        address_reader : `AddressReader`
            Reader object that can be used to look up UUIDs from integer
            indexes.
        empty_data_id : `lsst.daf.butler.DataCoordinate`
            The empty data ID for the appropriate dimension universe.

        Notes
        -----
        This method adds:

        - a ``bipartite_xgraph`` quantum node with full attributes;
        - ``bipartite_xgraph`` edges to adjacent datasets (which adds datasets
          nodes with no attributes), while populating those edge attributes;
        """
        task_init_node = graph.pipeline_graph.tasks[self.task_label].init
        graph._bipartite_xgraph.add_node(
            self.quantum_id, data_id=empty_data_id, task_label=self.task_label, pipeline_node=task_init_node
        )
        for connection_name, dataset_index in self.inputs.items():
            read_edge = task_init_node.get_input_edge(connection_name)
            dataset_id = address_reader.find(dataset_index).key
            graph._bipartite_xgraph.add_edge(dataset_id, self.quantum_id, is_read=True)
            graph._bipartite_xgraph.edges[dataset_id, self.quantum_id].setdefault(
                "pipeline_edges", []
            ).append(read_edge)
        for connection_name, dataset_index in self.outputs.items():
            write_edge = task_init_node.get_output_edge(connection_name)
            dataset_id = address_reader.find(dataset_index).key
            graph._bipartite_xgraph.add_edge(
                self.quantum_id,
                dataset_id,
                is_read=False,
                # There can only be one pipeline edge for an output.
                pipeline_edges=[write_edge],
            )
        graph._init_quanta[self.task_label] = self.quantum_id

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class ProvenanceInitQuantaModel(pydantic.RootModel):
    """Data model for the init quanta in a provenance graph."""

    root: list[ProvenanceInitQuantumModel] = pydantic.Field(default_factory=list)
    """List of special "init" quanta, one for each task."""

    def _add_to_graph(self, graph: ProvenanceQuantumGraph, address_reader: AddressReader) -> None:
        """Add this quantum and its edges to datasets to a provenance graph.

        Parameters
        ----------
        graph : `ProvenanceQuantumGraph`
            Graph to update in place.
        address_reader : `AddressReader`
            Reader object that can be used to look up UUIDs from integer
            indexes.
        """
        empty_data_id = DataCoordinate.make_empty(graph.pipeline_graph.universe)
        for init_quantum in self.root:
            init_quantum._add_to_graph(graph, address_reader, empty_data_id=empty_data_id)

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class ProvenanceQuantumGraph(BaseQuantumGraph):
    """A quantum graph that represents processing that has already been
    executed.

    Parameters
    ----------
    header : `HeaderModel`
        General metadata shared with other quantum graph types.
    pipeline_graph : `.pipeline_graph.PipelineGraph`
        Graph of tasks and dataset types.  May contain a superset of the tasks
        and dataset types that actually have quanta and datasets in the quantum
        graph.

    Notes
    -----
    A provenance quantum graph is generally obtained via the
    `ProvenanceQuantumGraphReader.graph` attribute, which is updated in-place
    as information is read from disk.
    """

    def __init__(self, header: HeaderModel, pipeline_graph: PipelineGraph) -> None:
        super().__init__(header, pipeline_graph)
        self._init_quanta: dict[TaskLabel, uuid.UUID] = {}
        self._quantum_only_xgraph = networkx.DiGraph()
        self._bipartite_xgraph = networkx.DiGraph()
        self._quanta_by_task_label: dict[str, dict[DataCoordinate, uuid.UUID]] = {
            task_label: {} for task_label in self.pipeline_graph.tasks.keys()
        }
        self._datasets_by_type: dict[str, dict[DataCoordinate, uuid.UUID]] = {
            dataset_type_name: {} for dataset_type_name in self.pipeline_graph.dataset_types.keys()
        }

    @property
    def init_quanta(self) -> Mapping[TaskLabel, uuid.UUID]:
        """A mapping from task label to the ID of the special init quantum for
        that task.

        This is populated by the ``init_quanta`` component.  Additional
        information about each init quantum can be found by using the ID to
        look up node attributes in the `bipartite_xgraph`, i.e.::

            info: ProvenanceInitQuantumInfo = qg.bipartite_xgraph.nodes[id]
        """
        return self._init_quanta

    @property
    def quanta_by_task(self) -> Mapping[TaskLabel, Mapping[DataCoordinate, uuid.UUID]]:
        """A nested mapping of all quanta, keyed first by task name and then by
        data ID.

        Notes
        -----
        This is populated one quantum at a time as they are read.  All tasks in
        the pipeline graph are included, even if none of their quanta were
        loaded (i.e. nested mappings may be empty).

        The returned object may be an internal dictionary; as the type
        annotation indicates, it should not be modified in place.
        """
        return self._quanta_by_task_label

    @property
    def datasets_by_type(self) -> Mapping[DatasetTypeName, Mapping[DataCoordinate, uuid.UUID]]:
        """A nested mapping of all datasets, keyed first by dataset type name
        and then by data ID.

        Notes
        -----
        This is populated one dataset at a time as they are read. All dataset
        types in the pipeline graph are included, even if none of their
        datasets were loaded (i.e. nested mappings may be empty).

        The returned object may be an internal dictionary; as the type
        annotation indicates, it should not be modified in place.
        """
        return self._datasets_by_type

    @property
    def quantum_only_xgraph(self) -> networkx.DiGraph:
        """A directed acyclic graph with quanta as nodes (and datasets elided).

        Notes
        -----
        Node keys are quantum UUIDs, and are populated one quantum at a time as
        they are loaded.  Loading quanta (via
        `ProvenanceQuantumGraphReader.read_quanta`) will add the loaded nodes
        with full attributes and add edges to adjacent nodes with no
        attributes. Loading datasets (via
        `ProvenanceQuantumGraphReader.read_datasets`) will also add edges and
        nodes with no attributes.

        Node attributes are described by the `ProvenanceQuantumInfo` types.

        This graph does not include special "init" quanta.

        The returned object is a read-only view of an internal one.
        """
        return self._quantum_only_xgraph.copy(as_view=True)

    @property
    def bipartite_xgraph(self) -> networkx.DiGraph:
        """A directed acyclic graph with quantum and dataset nodes.

        Notes
        -----
        Node keys are quantum or dataset UUIDs, and are populated one quantum
        or dataset at a time as they are loaded.  Loading quanta (via
        `ProvenanceQuantumGraphReader.read_quanta`) or datasets (via
        `ProvenanceQuantumGraphReader.read_datasets`) will load those nodes
        with full attributes and edges to adjacent nodes with no attributes.
        Loading quanta necessary to populate edge attributes.

        Node attributes are described by the
        `ProvenanceQuantumInfo`, `ProvenanceInitQuantumInfo`, and
        `ProvenanceDatasetInfo` types.

        This graph includes init-input and init-output datasets, but it does
        *not* reflect the dependency between each task's special "init" quantum
        and its runtime quanta (as this would require edges between quanta, and
        that would break the "bipartite" property).

        The returned object is a read-only view of an internal one.
        """
        return self._bipartite_xgraph.copy(as_view=True)

    def make_quantum_table(self) -> astropy.table.Table:
        """Construct an `astropy.table.Table` with a tabular summary of the
        quanta.

        Returns
        -------
        table : `astropy.table.Table`
            A table view of the quantum information.  This only includes
            counts of status categories and caveats, not any per-data-ID
            detail.

        Notes
        -----
        Success caveats in the table are represented by their
        `~QuantumSuccessCaveats.concise` form, so when pretty-printing this
        table for users, the `~QuantumSuccessCaveats.legend` should generally
        be printed as well.
        """
        rows = []
        for task_label, quanta_for_task in self.quanta_by_task.items():
            if not self.header.n_task_quanta[task_label]:
                continue
            status_counts = Counter[QuantumRunStatus](
                self._quantum_only_xgraph.nodes[q]["status"] for q in quanta_for_task.values()
            )
            caveat_counts = Counter[QuantumSuccessCaveats | None](
                self._quantum_only_xgraph.nodes[q]["caveats"] for q in quanta_for_task.values()
            )
            caveat_counts.pop(QuantumSuccessCaveats.NO_CAVEATS, None)
            caveat_counts.pop(None, None)
            if len(caveat_counts) > 1:
                caveats = "(multiple)"
            elif len(caveat_counts) == 1:
                ((code, count),) = caveat_counts.items()
                # MyPy can't tell that the pop(None, None) above makes None
                # impossible here.
                caveats = f"{code.concise()}({count})"  # type: ignore[union-attr]
            else:
                caveats = ""
            rows.append(
                {
                    "Task": task_label,
                    "Unknown": status_counts.get(QuantumRunStatus.METADATA_MISSING, 0),
                    "Successful": status_counts.get(QuantumRunStatus.SUCCESSFUL, 0),
                    "Caveats": caveats,
                    "Blocked": status_counts.get(QuantumRunStatus.BLOCKED, 0),
                    "Failed": status_counts.get(QuantumRunStatus.FAILED, 0),
                    "TOTAL": len(quanta_for_task),
                    "EXPECTED": self.header.n_task_quanta[task_label],
                }
            )
        return astropy.table.Table(rows)

    def make_exception_table(self) -> astropy.table.Table:
        """Construct an `astropy.table.Table` with counts for each exception
        type raised by each task.

        At present this only includes information from partial-outputs-error
        successes, since exception information for failures is not tracked.
        This may change in the future.

        Returns
        -------
        table : `astropy.table.Table`
            A table with columns for task label, exception type, and counts.
        """
        rows = []
        for task_label, quanta_for_task in self.quanta_by_task.items():
            counts_by_type = Counter(
                exc_info.type_name
                for q in quanta_for_task.values()
                if (exc_info := self._quantum_only_xgraph.nodes[q]["exception"]) is not None
            )
            for type_name, count in counts_by_type.items():
                rows.append({"Task": task_label, "Exception": type_name, "Count": count})
        return astropy.table.Table(rows)

    def make_task_resource_usage_table(
        self, task_label: TaskLabel, include_data_ids: bool = False
    ) -> astropy.table.Table:
        """Make a table of resource usage for a single task.

        Parameters
        ----------
        task_label : `str`
            Label of the task to extract resource usage for.
        include_data_ids : `bool`, optional
            Whether to also include data ID columns.

        Returns
        -------
        table : `astropy.table.Table`
            A table with columns for quantum ID and all fields in
            `QuantumResourceUsage`.
        """
        quanta_for_task = self.quanta_by_task[task_label]
        dtype_terms: list[tuple[str, np.dtype]] = [("quantum_id", np.dtype((np.void, 16)))]
        if include_data_ids:
            dimensions = self.pipeline_graph.tasks[task_label].dimensions
            for dimension_name in dimensions.data_coordinate_keys:
                dtype = np.dtype(self.pipeline_graph.universe.dimensions[dimension_name].primary_key.pytype)
                dtype_terms.append((dimension_name, dtype))
        fields = QuantumResourceUsage.get_numpy_fields()
        dtype_terms.extend(fields.items())
        row_dtype = np.dtype(dtype_terms)
        rows: list[object] = []
        for data_id, quantum_id in quanta_for_task.items():
            info: ProvenanceQuantumInfo = self._quantum_only_xgraph.nodes[quantum_id]
            if (resource_usage := info["resource_usage"]) is not None:
                row: tuple[object, ...] = (quantum_id.bytes,)
                if include_data_ids:
                    row += data_id.full_values
                row += resource_usage.get_numpy_row()
                rows.append(row)
        array = np.array(rows, dtype=row_dtype)
        return astropy.table.Table(array, units=QuantumResourceUsage.get_units())


@dataclasses.dataclass
class ProvenanceQuantumGraphReader(BaseQuantumGraphReader):
    """A helper class for reading provenance quantum graphs.

    Notes
    -----
    The `open` context manager should be used to construct new instances.
    Instances cannot be used after the context manager exits, except to access
    the `graph` attribute`.

    The various ``read_*`` methods in this class update the `graph` attribute
    in place and return ``self``.
    """

    graph: ProvenanceQuantumGraph = dataclasses.field(init=False)
    """Loaded provenance graph, populated in place as components are read."""

    @classmethod
    @contextmanager
    def open(
        cls,
        uri: ResourcePathExpression,
        *,
        page_size: int | None = None,
        import_mode: TaskImportMode = TaskImportMode.DO_NOT_IMPORT,
    ) -> Iterator[ProvenanceQuantumGraphReader]:
        """Construct a reader from a URI.

        Parameters
        ----------
        uri : convertible to `lsst.resources.ResourcePath`
            URI to open.  Should have a ``.qg`` extension.
        page_size : `int`, optional
            Approximate number of bytes to read at once from address files and
            multi-block files. Note that this does not set a page size for
            *all* reads, but it does affect the smallest, most numerous reads.
            Can also be set via the ``LSST_QG_PAGE_SIZE`` environment variable.
        import_mode : `.pipeline_graph.TaskImportMode`, optional
            How to handle importing the task classes referenced in the pipeline
            graph.

        Returns
        -------
        reader : `contextlib.AbstractContextManager` [ \
                `ProvenanceQuantumGraphReader` ]
            A context manager that returns the reader when entered.
        """
        with cls._open(
            uri,
            graph_type="provenance",
            address_filename="nodes",
            page_size=page_size,
            import_mode=import_mode,
            n_addresses=4,
        ) as self:
            yield self

    def __post_init__(self) -> None:
        self.graph = ProvenanceQuantumGraph(self.header, self.pipeline_graph)

    def read_init_quanta(self) -> Self:
        """Read the thin graph, with all edge information and categorization of
        quanta by task label.

        Returns
        -------
        self : `ProvenanceQuantumGraphReader`
            The reader (to permit method-chaining).
        """
        init_quanta = self._read_single_block("init_quanta", ProvenanceInitQuantaModel)
        for init_quantum in init_quanta.root:
            self.graph._init_quanta[init_quantum.task_label] = init_quantum.quantum_id
        init_quanta._add_to_graph(self.graph, self.address_reader)
        return self

    def read_full_graph(self) -> Self:
        """Read all bipartite edges and all quantum and dataset node
        attributes, fully populating the `graph` attribute.

        Returns
        -------
        self : `ProvenanceQuantumGraphReader`
            The reader (to permit method-chaining).

        Notes
        -----
        This does not read logs, metadata, or packages ; those must always be
        fetched explicitly.
        """
        self.read_init_quanta()
        self.read_datasets()
        self.read_quanta()
        return self

    def read_datasets(self, datasets: Iterable[uuid.UUID | DatasetIndex] | None = None) -> Self:
        """Read information about the given datasets.

        Parameters
        ----------
        datasets : `~collections.abc.Iterable` [`uuid.UUID` or `int`], optional
            Iterable of dataset IDs or indices to load.  If not provided, all
            datasets will be loaded.  The UUIDs and indices of quanta will be
            ignored.

        Return
        -------
        self : `ProvenanceQuantumGraphReader`
            The reader (to permit method-chaining).
        """
        return self._read_nodes(datasets, DATASET_ADDRESS_INDEX, DATASET_MB_NAME, ProvenanceDatasetModel)

    def read_quanta(self, quanta: Iterable[uuid.UUID | QuantumIndex] | None = None) -> Self:
        """Read information about the given quanta.

        Parameters
        ----------
        quanta : `~collections.abc.Iterable` [`uuid.UUID` or `int`], optional
            Iterable of quantum IDs or indices to load.  If not provided, all
            quanta will be loaded.  The UUIDs and indices of datasets and
            special init quanta will be ignored.

        Return
        -------
        self : `ProvenanceQuantumGraphReader`
            The reader (to permit method-chaining).
        """
        return self._read_nodes(quanta, QUANTUM_ADDRESS_INDEX, QUANTUM_MB_NAME, ProvenanceQuantumModel)

    def _read_nodes(
        self,
        nodes: Iterable[uuid.UUID | int] | None,
        address_index: int,
        mb_name: str,
        model_type: type[ProvenanceDatasetModel] | type[ProvenanceQuantumModel],
    ) -> Self:
        node: ProvenanceDatasetModel | ProvenanceQuantumModel | None
        if nodes is None:
            self.address_reader.read_all()
            nodes = self.address_reader.rows.keys()
            for node in MultiblockReader.read_all_models_in_zip(
                self.zf,
                mb_name,
                model_type,
                self.decompressor,
                int_size=self.header.int_size,
                page_size=self.page_size,
            ):
                if "pipeline_node" in self.graph._bipartite_xgraph.nodes.get(node.node_id, {}):
                    # Use the old node to reduce memory usage (since it might
                    # also have other outstanding reference holders).
                    continue
                node._add_to_graph(self.graph, self.address_reader)
        with MultiblockReader.open_in_zip(self.zf, mb_name, int_size=self.header.int_size) as mb_reader:
            for node_id_or_index in nodes:
                address_row = self.address_reader.find(node_id_or_index)
                if "pipeline_node" in self.graph._bipartite_xgraph.nodes.get(address_row.key, {}):
                    # Use the old node to reduce memory usage (since it might
                    # also have other outstanding reference holders).
                    continue
                node = mb_reader.read_model(
                    address_row.addresses[address_index], model_type, self.decompressor
                )
                if node is not None:
                    node._add_to_graph(self.graph, self.address_reader)
        return self

    def fetch_logs(
        self, nodes: Iterable[uuid.UUID | DatasetIndex | QuantumIndex]
    ) -> dict[uuid.UUID | DatasetIndex | QuantumIndex, ButlerLogRecords]:
        """Fetch log datasets.

        Parameters
        ----------
        nodes : `~collections.abc.Iterable` [ `uuid.UUID` ]
            UUIDs of the log datasets themselves or of the quanta they
            correspond to.

        Returns
        -------
        logs : `dict` [ `uuid.UUID`, `ButlerLogRecords`]
            Logs for the given IDs.
        """
        from lsst.daf.butler.logging import ButlerLogRecords

        result: dict[uuid.UUID | DatasetIndex | QuantumIndex, ButlerLogRecords] = {}
        with MultiblockReader.open_in_zip(self.zf, LOG_MB_NAME, int_size=self.header.int_size) as mb_reader:
            for node_id_or_index in nodes:
                address_row = self.address_reader.find(node_id_or_index)
                log = mb_reader.read_model(
                    address_row.addresses[LOG_ADDRESS_INDEX], ButlerLogRecords, self.decompressor
                )
                if log is not None:
                    result[node_id_or_index] = log
        return result

    def fetch_metadata(
        self, nodes: Iterable[uuid.UUID | DatasetIndex | QuantumIndex]
    ) -> dict[uuid.UUID | DatasetIndex | QuantumIndex, TaskMetadata]:
        """Fetch metadata datasets.

        Parameters
        ----------
        nodes : `~collections.abc.Iterable` [ `uuid.UUID` ]
            UUIDs of the metadata datasets themselves or of the quanta they
            correspond to.

        Returns
        -------
        metadata : `dict` [ `uuid.UUID`, `TaskMetadata`]
            Metadata for the given IDs.
        """
        from .._task_metadata import TaskMetadata

        result: dict[uuid.UUID | DatasetIndex | QuantumIndex, TaskMetadata] = {}
        with MultiblockReader.open_in_zip(
            self.zf, METADATA_MB_NAME, int_size=self.header.int_size
        ) as mb_reader:
            for node_id_or_index in nodes:
                address_row = self.address_reader.find(node_id_or_index)
                metadata = mb_reader.read_model(
                    address_row.addresses[METADATA_ADDRESS_INDEX], TaskMetadata, self.decompressor
                )
                if metadata is not None:
                    result[node_id_or_index] = metadata
        return result

    def fetch_packages(self) -> Packages:
        """Fetch package version information."""
        data = self._read_single_block_raw("packages")
        return Packages.fromBytes(data, format="json")



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Scanner",)

import dataclasses
import itertools
import uuid

import zstandard

from lsst.daf.butler import ButlerLogRecords, DatasetRef, QuantumBackedButler
from lsst.utils.iteration import ensure_iterable

from ... import automatic_connection_constants as acc
from ..._status import QuantumSuccessCaveats
from ..._task_metadata import TaskMetadata
from ...pipeline_graph import PipelineGraph, TaskImportMode
from ...quantum_provenance_graph import ExceptionInfo
from ...resource_usage import QuantumResourceUsage
from .._multiblock import Compressor
from .._predicted import (
    PredictedDatasetModel,
    PredictedQuantumDatasetsModel,
    PredictedQuantumGraphReader,
)
from ._communicators import ScannerCommunicator
from ._structs import IngestRequest, ScanReport, ScanResult, ScanStatus


@dataclasses.dataclass
class Scanner:
    """A helper class for the provenance aggregator that reads metadata and log
    files and scans for which outputs exist.
    """

    predicted_path: str
    """Path to the predicted quantum graph."""

    butler_path: str
    """Path or alias to the central butler repository."""

    comms: ScannerCommunicator
    """Communicator object for this worker."""

    reader: PredictedQuantumGraphReader = dataclasses.field(init=False)
    """Reader for the predicted quantum graph."""

    qbb: QuantumBackedButler = dataclasses.field(init=False)
    """A quantum-backed butler used for log and metadata reads and existence
    checks for other outputs (when necessary).
    """

    compressor: Compressor | None = None
    """Object used to compress JSON blocks.

    This is `None` until a compression dictionary is received from the writer
    process.
    """

    init_quanta: dict[uuid.UUID, PredictedQuantumDatasetsModel] = dataclasses.field(init=False)
    """Dictionary mapping init quantum IDs to their predicted models."""

    def __post_init__(self) -> None:
        if self.comms.config.mock_storage_classes:
            import lsst.pipe.base.tests.mocks  # noqa: F401
        self.comms.log.verbose("Reading from predicted quantum graph.")
        self.reader = self.comms.enter(
            PredictedQuantumGraphReader.open(self.predicted_path, import_mode=TaskImportMode.DO_NOT_IMPORT)
        )
        self.reader.read_dimension_data()
        self.reader.read_init_quanta()
        self.comms.log.verbose("Initializing quantum-backed butler.")
        self.qbb = self.make_qbb(self.butler_path, self.reader.pipeline_graph)
        self.init_quanta = {q.quantum_id: q for q in self.reader.components.init_quanta.root}

    @staticmethod
    def make_qbb(butler_config: str, pipeline_graph: PipelineGraph) -> QuantumBackedButler:
        """Make quantum-backed butler that can operate on the outputs of the
        quantum graph.

        Parameters
        ----------
        butler_config : `str`
            Path or alias for the central butler repository that shares storage
            with the quantum-backed butler.
        pipeline_graph : `..pipeline_graph.PipelineGraph`
            Graph of tasks and dataset types.

        Returns
        -------
        qbb : `lsst.daf.butler.QuantumBackedButler`
            Quantum-backed butler.  This does not have the datastore records
            needed to read overall-inputs.
        """
        return QuantumBackedButler.from_predicted(
            butler_config,
            predicted_inputs=[],
            predicted_outputs=[],
            dimensions=pipeline_graph.universe,
            # We don't need the datastore records in the QG because we're
            # only going to read metadata and logs, and those are never
            # overall inputs.
            datastore_records={},
            dataset_types={node.name: node.dataset_type for node in pipeline_graph.dataset_types.values()},
        )

    @property
    def pipeline_graph(self) -> PipelineGraph:
        """Graph of tasks and dataset types."""
        return self.reader.pipeline_graph

    @staticmethod
    def run(predicted_path: str, butler_path: str, comms: ScannerCommunicator) -> None:
        """Run the scanner.

        Parameters
        ----------
        predicted_path : `str`
            Path to the predicted quantum graph.
        butler_path : `str`
            Path or alias to the central butler repository.
        comms : `ScannerCommunicator`
            Communicator for the scanner.

        Notes
        -----
        This method is designed to run as the ``target`` in
        `WorkerContext.make_worker`.
        """
        with comms:
            scanner = Scanner(predicted_path, butler_path, comms)
            scanner.loop()

    def loop(self) -> None:
        """Run the main loop for the scanner."""
        self.comms.log.info("Scan request loop beginning.")
        for quantum_id in self.comms.poll():
            if self.compressor is None and (cdict_data := self.comms.get_compression_dict()) is not None:
                self.compressor = zstandard.ZstdCompressor(
                    self.comms.config.zstd_level, zstandard.ZstdCompressionDict(cdict_data)
                )
            self.scan_quantum(quantum_id)

    def scan_dataset(self, predicted: PredictedDatasetModel) -> bool:
        """Scan for a dataset's existence.

        Parameters
        ----------
        predicted : `.PredictedDatasetModel`
            Information about the dataset from the predicted graph.

        Returns
        -------
        exists : `bool``
            Whether the dataset exists
        """
        ref = self.reader.components.make_dataset_ref(predicted)
        return self.qbb.stored(ref)

    def scan_quantum(self, quantum_id: uuid.UUID) -> ScanResult:
        """Scan for a quantum's completion and error status, and its output
        datasets' existence.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            Unique ID for the quantum.

        Returns
        -------
        result : `ScanResult`
            Scan result struct.
        """
        if (predicted_quantum := self.init_quanta.get(quantum_id)) is not None:
            result = ScanResult(predicted_quantum.quantum_id, status=ScanStatus.INIT)
            self.comms.log.debug("Created init scan for %s (%s)", quantum_id, predicted_quantum.task_label)
        else:
            self.reader.read_quantum_datasets([quantum_id])
            predicted_quantum = self.reader.components.quantum_datasets[quantum_id]
            self.comms.log.debug(
                "Scanning %s (%s@%s)",
                quantum_id,
                predicted_quantum.task_label,
                predicted_quantum.data_coordinate,
            )
            result = ScanResult(predicted_quantum.quantum_id, ScanStatus.INCOMPLETE)
            del self.reader.components.quantum_datasets[quantum_id]
            log_id = self._read_and_compress_log(predicted_quantum, result)
            if not self.comms.config.assume_complete and not result.log:
                self.comms.log.debug("Abandoning scan for %s; no log dataset.", quantum_id)
                result.status = ScanStatus.ABANDONED
                self.comms.report_scan(ScanReport(result.quantum_id, result.status))
                return result
            metadata_id = self._read_and_compress_metadata(predicted_quantum, result)
            if result.metadata:
                result.status = ScanStatus.SUCCESSFUL
                result.existing_outputs.add(metadata_id)
            elif self.comms.config.assume_complete:
                result.status = ScanStatus.FAILED
            else:
                # We found the log dataset, but no metadata; this means the
                # quantum failed, but a retry might still happen that could
                # turn it into a success if we can't yet assume the run is
                # complete.
                self.comms.log.debug("Abandoning scan for %s.", quantum_id)
                result.status = ScanStatus.ABANDONED
                self.comms.report_scan(ScanReport(result.quantum_id, result.status))
                return result
            if result.log:
                result.existing_outputs.add(log_id)
        for predicted_output in itertools.chain.from_iterable(predicted_quantum.outputs.values()):
            if predicted_output.dataset_id not in result.existing_outputs and self.scan_dataset(
                predicted_output
            ):
                result.existing_outputs.add(predicted_output.dataset_id)
        to_ingest = self._make_ingest_request(predicted_quantum, result)
        self.comms.report_scan(ScanReport(result.quantum_id, result.status))
        assert result.status is not ScanStatus.INCOMPLETE
        assert result.status is not ScanStatus.ABANDONED
        if self.comms.config.output_path is not None:
            self.comms.request_write(result)
        self.comms.request_ingest(to_ingest)
        self.comms.log.debug("Finished scan for %s.", quantum_id)
        return result

    def _make_ingest_request(
        self, predicted_quantum: PredictedQuantumDatasetsModel, result: ScanResult
    ) -> IngestRequest:
        """Make an ingest request from a quantum scan.

        Parameters
        ----------
        predicted_quantum : `PredictedQuantumDatasetsModel`
            Information about the predicted quantum.
        result : `ScanResult`
            Result of a quantum scan.

        Returns
        -------
        ingest_request : `IngestRequest`
            A request to be sent to the ingester.
        """
        predicted_outputs_by_id = {
            d.dataset_id: d for d in itertools.chain.from_iterable(predicted_quantum.outputs.values())
        }
        to_ingest_predicted: list[PredictedDatasetModel] = []
        to_ingest_refs: list[DatasetRef] = []
        for dataset_id in result.existing_outputs:
            predicted_output = predicted_outputs_by_id[dataset_id]
            to_ingest_predicted.append(predicted_output)
            to_ingest_refs.append(self.reader.components.make_dataset_ref(predicted_output))
        to_ingest_records = self.qbb._datastore.export_predicted_records(to_ingest_refs)
        return IngestRequest(result.quantum_id, to_ingest_predicted, to_ingest_records)

    def _read_and_compress_metadata(
        self, predicted_quantum: PredictedQuantumDatasetsModel, result: ScanResult
    ) -> uuid.UUID:
        """Attempt to read the metadata dataset for a quantum to extract
        provenance information from it.

        Parameters
        ----------
        predicted_quantum : `PredictedQuantumDatasetsModel`
            Information about the predicted quantum.
        result : `ScanResult`
            Result object to be modified in-place.

        Returns
        -------
        dataset_id : `uuid.UUID`
            UUID of the metadata dataset.
        """
        assert not result.metadata, "We shouldn't be scanning again if we already read the metadata."
        (predicted_dataset,) = predicted_quantum.outputs[acc.METADATA_OUTPUT_CONNECTION_NAME]
        ref = self.reader.components.make_dataset_ref(predicted_dataset)
        try:
            # This assumes QBB metadata writes are atomic, which should be the
            # case. If it's not we'll probably get pydantic validation errors
            # here.
            content: TaskMetadata = self.qbb.get(ref, storageClass="TaskMetadata")
        except FileNotFoundError:
            if not self.comms.config.assume_complete:
                return ref.id
        else:
            try:
                # Int conversion guards against spurious conversion to
                # float that can apparently sometimes happen in
                # TaskMetadata.
                result.caveats = QuantumSuccessCaveats(int(content["quantum"]["caveats"]))
            except LookupError:
                pass
            try:
                result.exception = ExceptionInfo._from_metadata(
                    content[predicted_quantum.task_label]["failure"]
                )
            except LookupError:
                pass
            try:
                result.existing_outputs = {
                    uuid.UUID(id_str) for id_str in ensure_iterable(content["quantum"].getArray("outputs"))
                }
            except LookupError:
                pass
            result.resource_usage = QuantumResourceUsage.from_task_metadata(content)
            result.metadata = content.model_dump_json().encode()
            if self.compressor is not None:
                result.metadata = self.compressor.compress(result.metadata)
                result.is_compressed = True
        return ref.id

    def _read_and_compress_log(
        self, predicted_quantum: PredictedQuantumDatasetsModel, result: ScanResult
    ) -> uuid.UUID:
        """Attempt to read the log dataset for a quantum to test for the
        quantum's completion (the log is always written last) and aggregate
        the log content in the provenance quantum graph.

        Parameters
        ----------
        predicted_quantum : `PredictedQuantumDatasetsModel`
            Information about the predicted quantum.
        result : `ScanResult`
            Result object to be modified in-place.

        Returns
        -------
        dataset_id : `uuid.UUID`
            UUID of the log dataset.
        """
        (predicted_dataset,) = predicted_quantum.outputs[acc.LOG_OUTPUT_CONNECTION_NAME]
        ref = self.reader.components.make_dataset_ref(predicted_dataset)
        try:
            # This assumes QBB log writes are atomic, which should be the case.
            # If it's not we'll probably get pydantic validation errors here.
            content: ButlerLogRecords = self.qbb.get(ref)
        except FileNotFoundError:
            if not self.comms.config.assume_complete:
                return ref.id
        else:
            result.log = content.model_dump_json().encode()
            if self.compressor is not None:
                result.log = self.compressor.compress(result.log)
                result.is_compressed = True
        return ref.id



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "IngestRequest",
    "ScanReport",
    "ScanResult",
    "ScanStatus",
)

import dataclasses
import enum
import uuid

from lsst.daf.butler.datastore.record_data import DatastoreRecordData

from ..._status import QuantumSuccessCaveats
from ...quantum_provenance_graph import ExceptionInfo, QuantumRunStatus
from ...resource_usage import QuantumResourceUsage
from .._common import DatastoreName
from .._predicted import PredictedDatasetModel


class ScanStatus(enum.Enum):
    """Status enum for quantum scanning.

    Note that this records the status for the *scanning* which is distinct
    from the status of the quantum's execution.
    """

    INCOMPLETE = enum.auto()
    """The quantum is not necessarily done running, and cannot be scanned
    conclusively yet.
    """

    ABANDONED = enum.auto()
    """The quantum's execution appears to have failed but we cannot rule out
    the possibility that it could be recovered, but we've also waited long
    enough (according to `ScannerTimeConfigDict.retry_timeout`) that it's time
    to stop trying for now.

    This state means a later run with `ScannerConfig.assume_complete` is
    required.
    """

    SUCCESSFUL = enum.auto()
    """The quantum was conclusively scanned and was executed successfully,
    unblocking scans for downstream quanta.
    """

    FAILED = enum.auto()
    """The quantum was conclusively scanned and failed execution, blocking
    scans for downstream quanta.
    """

    BLOCKED = enum.auto()
    """A quantum upstream of this one failed."""

    INIT = enum.auto()
    """Init quanta need special handling, because they don't have logs and
    metadata.
    """


@dataclasses.dataclass
class ScanReport:
    """Minimal information needed about a completed scan by the supervisor."""

    quantum_id: uuid.UUID
    """Unique ID of the quantum."""

    status: ScanStatus
    """Combined status of the scan and the execution of the quantum."""


@dataclasses.dataclass
class IngestRequest:
    """A request to ingest datasets produced by a single quantum."""

    producer_id: uuid.UUID
    """ID of the quantum that produced these datasets."""

    datasets: list[PredictedDatasetModel]
    """Registry information about the datasets."""

    records: dict[DatastoreName, DatastoreRecordData]
    """Datastore information about the datasets."""

    def __bool__(self) -> bool:
        return bool(self.datasets or self.records)


@dataclasses.dataclass
class ScanResult:
    """A struct that represents the result of scanning a quantum."""

    quantum_id: uuid.UUID
    """Unique ID for the quantum."""

    status: ScanStatus
    """Combined status for the scan and the execution of the quantum."""

    caveats: QuantumSuccessCaveats | None = None
    """Flags indicating caveats on successful quanta."""

    exception: ExceptionInfo | None = None
    """Information about an exception raised when the quantum was executing."""

    resource_usage: QuantumResourceUsage | None = None
    """Resource usage information (timing, memory use) for this quantum."""

    existing_outputs: set[uuid.UUID] = dataclasses.field(default_factory=set)
    """Unique IDs of the output datasets that were actually written."""

    metadata: bytes = b""
    """Raw content of the metadata dataset."""

    log: bytes = b""
    """Raw content of the log dataset."""

    is_compressed: bool = False
    """Whether the `metadata` and `log` attributes are compressed."""

    def get_run_status(self) -> QuantumRunStatus:
        """Translate the scan status and metadata/log presence into a run
        status.
        """
        if self.status is ScanStatus.BLOCKED:
            return QuantumRunStatus.BLOCKED
        if self.status is ScanStatus.INIT:
            return QuantumRunStatus.SUCCESSFUL
        if self.log:
            if self.metadata:
                return QuantumRunStatus.SUCCESSFUL
            else:
                return QuantumRunStatus.FAILED
        else:
            if self.metadata:
                return QuantumRunStatus.LOGS_MISSING
            else:
                return QuantumRunStatus.METADATA_MISSING
